using System;
using System.Windows.Forms;
using System.Collections;
using System.IO;
using System.Xml;

namespace UserChat
{
	/// <summary>
	/// Summary description for NetworkStatus.
	/// </summary>
	public class NetworkStatus
	{
		Hashtable hashNetwork;
		ArrayList visited;
		//XmlDoc accessible from anywhere
		private XmlDocument xmlDoc;
		private Hashtable neighbors_hash;
		private NT_CONSTANTS NTConstants;

		public NetworkStatus()
		{
			hashNetwork = new Hashtable();
			xmlDoc = new XmlDocument();
			xmlDoc.Load(new StreamReader("\\My Documents\\Personal\\conf.xml"));
			neighbors_hash = new Hashtable();
			visited = new ArrayList();
			NTConstants = new NT_CONSTANTS();
		}

		public NetworkStatus(Hashtable p_hashNetwork)
		{
			hashNetwork = new Hashtable();
			hashNetwork = p_hashNetwork;
			xmlDoc = new XmlDocument();
			xmlDoc.Load(new StreamReader("\\My Documents\\Personal\\conf.xml"));
			neighbors_hash = new Hashtable();
			visited = new ArrayList();
			NTConstants = new NT_CONSTANTS();
		}

		private double getDistance(Point p1, Point p2)
		{
			double powDistance = Math.Pow((p1.row-p2.row),2)+
				Math.Pow((p1.col-p2.col),2);
			double distance = Math.Sqrt(powDistance); 

			return Convert.ToDouble(Math.Round(Convert.ToDecimal(distance),2));
		}
		
		public bool isNeighbor(Point p1, Point p2)
		{
			if(getDistance(p1,p2)<=NTConstants.POWER_RANGE)
			{
				return true;
			}
			else
			{
				return false;   
			}
		}//isNeighbor()

		private Point getStartPoint(string p_source_id)
		{
			try
			{
				XmlNode root = xmlDoc.DocumentElement;//configuration root element
				XmlNodeList xNodes = root.ChildNodes;
											
				foreach(XmlNode childNode in xNodes)
				{
					
					XmlElement currentElement = (XmlElement)childNode;

					if(currentElement.GetAttribute("ID").Equals(p_source_id))
					{
						string stPt = currentElement.GetAttribute("StartPt");
						double row = getRow(stPt);
						double col = getCol(stPt);
						Point startPoint = new Point(row, col);
						return startPoint;
					}
				}		
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception occurred in getStartPoint() : "+e.Message);
			}
			return new Point(-1,-1);           
		}
		/// <summary>
		/// When string (row,col) is passed as a parameter, return row as an integer
		/// </summary>
		/// <param name="s">destination</param>
		/// <returns>row as an integer</returns>
		public int getRow(string s)
		{
			try
			{
				int leftP = s.IndexOf("(");		
				int dot = s.IndexOf(",");
				string strRow = s.Substring(leftP+1, dot-1);
				//MessageBox.Show("strRow : "+strRow);
				int row = Convert.ToInt32(strRow);
				return row;
			}
			catch(Exception e)
			{
				MessageBox.Show("getRow() Exception occurred :"+e.Message);                
				return -1;
			}
		}

		/// <summary>
		/// When string (row,col) is passed as a parameter, return col as an integer
		/// </summary>
		/// <param name="s">destination</param>
		/// <returns>col as an integer</returns>
		public int getCol(string s)
		{
			try
			{
				int rightP = s.IndexOf(")");
				int dot = s.IndexOf(",");
				int left = dot+1;
				int right =rightP-left;
				//MessageBox.Show("left : "+left+" right :"+right);
				string strCol = s.Substring(dot+1, right);
				//MessageBox.Show("strCol : "+strCol);
				int col = Convert.ToInt32(strCol);
				return col;
			}
			catch(Exception e)
			{
				MessageBox.Show("getCol() Exception occurred :"+e.Message);                
				return -1;
			}
		}


		/// <summary>
		/// returns the exact point of location of each 
		/// node at a certain time T.
		/// </summary>
		/// <param name="T">time to wish to be checked</param>
		/// <param name="node_order">the index of node in all sequences[]</param>
		/// <returns>the exact point of location</returns>
		public Point getLocation(int T, string p_source)
		{
			ArrayList sequence = new ArrayList();
			IDictionaryEnumerator ide = hashNetwork.GetEnumerator();
			while(ide.MoveNext())
			{
				if(ide.Key.ToString().Equals(p_source))
				{
                   sequence = (ArrayList)ide.Value;
				}
			}
			Sequence[] rl = new Sequence[10];  
			sequence.CopyTo(0,rl,0,sequence.Count);
			double elapsed_time =0;			
			int i =0;
			Point startPoint = getStartPoint(p_source);
			double distance = 0;
			int speed = 0;
			int pause = 0;
			double previous_elapsed_time =0;

			double x1 = 0;
			double y1 = 0;
			double x2 = 0;
			double y2 = 0;
			
			while(elapsed_time<T)
			{
				if(i>=rl.Length)
				{
					i=0;
				}//when 10 destinations are traversed already, 
				//it circles by being back to the beginning.-->Reset

				if(i==0)
				{
					x1 = startPoint.row;
					y1 = startPoint.col;
				}
				else
				{
					x1 = rl[i-1].destination.row;
					y1 = rl[i-1].destination.col;
				}
				x2 = rl[i].destination.row;
				y2 = rl[i].destination.col;
				Point point1 = new Point (x1, y1);
				Point point2 = new Point (x2, y2);
				distance = getDistance(point1, point2);

				speed = rl[i].speed;
				pause = rl[i].pause;
				previous_elapsed_time = elapsed_time;
				elapsed_time = elapsed_time +(distance/speed);
				double remainedT =0;
				if(elapsed_time>T)//location is on the way to the destination
				{
					remainedT =T-previous_elapsed_time;
					double distanceFarFromPrevPoint = remainedT*speed; 
                     

					Point p = getPointOnTheWay(point1, point2,distanceFarFromPrevPoint);
					//MessageBox.Show("On the Way : "+p.row+","+p.col);
					return p;
				}
				else
				{
					elapsed_time = elapsed_time + pause;
					if(elapsed_time>=T)//point location stays at the destination point
					{
						Point p = new Point(rl[i].destination.row,rl[i].destination.col); 
						//MessageBox.Show("On the Destination : "+p.row+","+p.col);
						return p;                      
					}
				}
				i++;
			}//while
			if(T==0)
			{
				x1 = startPoint.row;
				y1 = startPoint.col;
				Point p = new Point(x1,y1);
				return p;
			}
			// MessageBox.Show("No Point (0,0)");
			return new Point(0,0);
		}//end of method
		private Point getPointOnTheWay(Point point1, Point point2, double p_distanceFarFromPoint1)
		{
			try
			{
				double x1 = Convert.ToDouble(Math.Round(Convert.ToDecimal(point1.row),2));
				double y1 = Convert.ToDouble(Math.Round(Convert.ToDecimal(point1.col),2));
				double x2 = Convert.ToDouble(Math.Round(Convert.ToDecimal(point2.row),2));
				double y2 = Convert.ToDouble(Math.Round(Convert.ToDecimal(point2.col),2));
				double distance = getDistance(point1, point2);
				double approximateDistance = Convert.ToDouble(Math.Round(Convert.ToDecimal(p_distanceFarFromPoint1),2));
				if(distance>p_distanceFarFromPoint1)				
				{
					double toCalculateY = Convert.ToDouble(Math.Round(Convert.ToDecimal(approximateDistance*(Math.Abs((y1-y2))/distance)),2));
					double toCalculateX = Convert.ToDouble(Math.Round(Convert.ToDecimal(approximateDistance*(Math.Abs((x1-x2))/distance)),2));
					if((x1-x2)*(y1-y2)==0)//when slope == zero
					{
						//MessageBox.Show("("+x1+","+y1+")  ("+x2+","+y2+")");
						//MessageBox.Show("when slope is zero");
						if((x1-x2)==0)
						{

							if(y1<y2)
							{
								double y = Convert.ToDouble(Math.Round(Convert.ToDecimal(y1+approximateDistance),2));
								Point p = new Point(x1, y);
								return p;
							}
							else if(y1>y2)
							{
								double y = Convert.ToDouble(Math.Round(Convert.ToDecimal(y1-approximateDistance),2));
								Point p = new Point(x1, y);
								return p;
							}
							else //if(y1==y2)
							{
								return point1;
							}
						}
						else//y1==y2
						{
							if(x1<x2)
							{
								double x = Convert.ToDouble(Math.Round(Convert.ToDecimal(x1+approximateDistance),2));
								Point p = new Point(x, y1);
								return p;                          
							}
							else if(x1>x2)
							{
								double x = Convert.ToDouble(Math.Round(Convert.ToDecimal(x1-approximateDistance),2));
								Point p = new Point(x, y1);
								return p;                          
							}
							else
							{
								return point1;
							}               
						}
					}
					else if((x1-x2)*(y1-y2)>0)//when slope >0
					{
						//MessageBox.Show("when slope > zero");
						//MessageBox.Show("("+x1+","+y1+")  ("+x2+","+y2+")");
						if(x1<x2)
						{
							Point p = new Point ((x1+toCalculateX), (y1+toCalculateY));
							return p;                      
						}
						else
						{
							Point p = new Point ((x1-toCalculateX), (y1-toCalculateY));
							return p;
						}
					}
					else//when slope < 0
					{
						//MessageBox.Show("when slope < zero");
						//MessageBox.Show("("+x1+","+y1+")  ("+x2+","+y2+")");
						if(x1>x2)
						{
							Point p = new Point ((x1-toCalculateX), (y1+toCalculateY));
							return p;                      

						}
						else //x1<x2
						{
							Point p = new Point ((x1+toCalculateX), (y1-toCalculateY));
							return p;                      
						}
					}
				}//if
				else
				{
					return new Point(-1,-1);
				}
			}//try
			catch(Exception e)
			{
				MessageBox.Show("Exception in getPointOnTheWay() : "+e.Message);
				Point p = new Point(-1, -1);
				return p;
			}
		
		}//end of method

		//for debugging
		public void showNodeList(Hashtable p_node_list)
		{
			try
			{
				IDictionaryEnumerator ide = p_node_list.GetEnumerator();
				string inside = "";
				while(ide.MoveNext())
				{ 
					Point point = (Point)ide.Value;
					inside = inside+ide.Key.ToString()+": ("+point.row.ToString()+" , "+
						point.col.ToString()+")";
					inside = inside+"\r\n";               
				}
				MessageBox.Show("Node List : "+inside);
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception in showNodeList(): "+e.Message);
			}
		}

		public int getNTConstants()
		{
             return NTConstants.POWER_RANGE;
		}

		public void adjustNetworkTopology(Hashtable p_node_list, bool p_OnACM, bool p_isStartPacket)
		{
			try
			{
				neighbors_hash.Clear();
				IDictionaryEnumerator outer_ide = p_node_list.GetEnumerator();
				
				IDictionaryEnumerator inner_ide = p_node_list.GetEnumerator();
 				while(outer_ide.MoveNext())
				{
					Point point1 = (Point)outer_ide.Value;
					ArrayList neighbors_al = new ArrayList();
				    					
					while(inner_ide.MoveNext())
					{					
						if(!outer_ide.Key.ToString().Equals(inner_ide.Key.ToString()))
						{
							Point point2 = (Point)inner_ide.Value;				
							if(isConnected(point1, point2))
							{
								neighbors_al.Add(inner_ide.Key.ToString());
							}
						}
					}//inner while
					neighbors_hash.Add(outer_ide.Key.ToString(), neighbors_al);
					inner_ide.Reset();//reset the inner ie
				}//outer while
//				MessageBox.Show("After collecting network neighbors : no modification");
//				showInsideNeighbors_hash();
				if(p_OnACM || p_isStartPacket)
				{
					modifyConnections(p_node_list);
				}
				//showInsideNeighbors_hash();
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in adjustNetworkTopology(): "+e.Message);
			}
		}

		public Hashtable getNeighbors_hash()
		{
            return neighbors_hash;
		}

		private void modifyConnections(Hashtable p_node_list)
		{
			try
			{
				Hashtable checkList = p_node_list;
				//string disconnectedNode ="NA";
				bool isAllConnectedWithMultiHops = false;
				checkMaxNeighbors(p_node_list);
				//MessageBox.Show("after checkMaxNeighbors()");
				//showInsideNeighbors_hash();
				int i = 0;
				int j = 0;
				string root = "";
				while(!isAllConnectedWithMultiHops)
				{
					ICollection ic = p_node_list.Keys;
					IEnumerator ie = ic.GetEnumerator();
					//showInsideNeighbors_hash();
					while(ie.MoveNext())
					{
						j++;
						if(j==1)
						{
							root = ie.Current.ToString(); 
						}
						string node = ie.Current.ToString();
						//      MessageBox.Show("Look for "+node+" in root "+root);
						//		showInsideNeighbors_hash();
						if(!isInSourceTree(root, node))
						{
							//MessageBox.Show("not isInSourceTree(): node "+node);
							getConnected(node, p_node_list);
							//MessageBox.Show("after getConnected()");
							//showInsideVisited();
							visited.Clear();
							ie.Reset();
							j = 0;
							i=0;
							break;
						}
						else
						{
							i++;
							if(i==ic.Count)
							{

								//				MessageBox.Show("All connected : i = "+i);
								isAllConnectedWithMultiHops = true;
								break;
							}
						}
					}               
				}
					
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in modifyConnections(): "+e.Message);
			}
		}

		private void getConnected(string p_node, Hashtable p_node_list)
		{
			try
			{
				IEnumerator ie = visited.GetEnumerator();
				string ProperNode = "";
				int minNeighbors = neighbors_hash.Count;
				while(ie.MoveNext())
				{
					string visitedNode =  ie.Current.ToString();
					//	MessageBox.Show("traversing visitedNode : "+visitedNode);
					ArrayList neighbors = getNeighbors(visitedNode);
					int neighborsNum = neighbors.Count;
					if(minNeighbors>neighborsNum)
					{
						minNeighbors = neighborsNum;
						ProperNode = visitedNode;
					}
				}//end of while
				//MessageBox.Show("selected ProperNode : "+ProperNode);
				makeNeighbors(ProperNode, p_node, p_node_list);
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getConnected():"+e.Message);
			}
		}

		private void makeNeighbors(string properNode, string p_node, Hashtable p_node_list)
		{
			try
			{
				IDictionaryEnumerator ide = p_node_list.GetEnumerator();
				int i = 0;
				while(ide.MoveNext())
				{
					if(ide.Key.ToString().Equals(properNode))
					{
						ArrayList al = getNeighbors(ide.Key.ToString());
						al.Add(p_node);
						neighbors_hash.Remove(ide.Key.ToString());
						neighbors_hash.Add(ide.Key.ToString(), al);
						i++;
					}
					else if(ide.Key.ToString().Equals(p_node))
					{
						ArrayList al = getNeighbors(ide.Key.ToString());
						al.Add(properNode);
						neighbors_hash.Remove(ide.Key.ToString());
						neighbors_hash.Add(ide.Key.ToString(), al);
						i++;
					}
					else if(i==2)
					{
						break;
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in makeNeighbors() : "+e.Message);
			}
		}

		private bool isInSourceTree(string root, string node)
		{
			try
			{
				//MessageBox.Show("Before isInSourceTree()");
				//showInsideNeighbors_hash();
				//showInsideVisited();
				bool isConnected = false;
				Queue Q = new Queue();
				Q.Enqueue(root);
				visited.Add(root);
				while(!(Q.Count==0))
				{
					string temp_node = Q.Dequeue().ToString();
					if(temp_node.Equals(node))
					{
						isConnected = true;
						visited.Clear();
						return isConnected;
					}
					else
					{
						ArrayList neighbors = getNeighbors(temp_node); 
						for(int i=0; i<neighbors.Count; i++)
						{							
							string nb = neighbors[i].ToString();
							if(!isVisited(nb))
							{
								//showInsideVisited();
								//MessageBox.Show("not visited node: "+nb);
								Q.Enqueue(nb);
								visited.Add(nb);
							}
						}
					}                   
				}//end of while
				return isConnected;
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in isInSourcTree() : "+e.Message);
				return false;
			}
		}

		private bool isVisited(string node)
		{
			if(visited.Contains(node))
			{
				return true;
			}
			else
			{
				return false;
			}
		}

		private bool isConnected(Point p1, Point p2)
		{
			if(getDistance(p1,p2)<=NTConstants.POWER_RANGE)
			{
				return true;
			}
			else
			{
				return false;   
			}
		}//is connected: bool
		/// <summary>
		/// Check if each node is directly connected with more than half of participating nodes. 
		/// If so, cut off the neighboring nodes which are located farther from the node.
		/// </summary>
		/// <param name="p_max_neighbors"></param>
		private void checkMaxNeighbors(Hashtable p_node_list)
		{
			try
			{
				int max_neighbors = neighbors_hash.Count/2;
				IDictionaryEnumerator ide = p_node_list.GetEnumerator();
				//MessageBox.Show("num of nodes : "+neighbors_hash.Count);
				while(ide.MoveNext())
				{
					ArrayList al = getNeighbors(ide.Key.ToString());
					int numOfNeighbors = al.Count;
					//	MessageBox.Show(ide.Key.ToString());
					//	MessageBox.Show("numOfNeighbors : "+numOfNeighbors);
					//	MessageBox.Show("max_neighbors : "+max_neighbors);

					while(numOfNeighbors>max_neighbors)
					{
						cutOffProperNode(ide.Key.ToString(), al, p_node_list);
						numOfNeighbors--;
					}
					//	MessageBox.Show("after cutting off numOfNeighbors : "+numOfNeighbors);

				}   
				//MessageBox.Show("checkMax 8");
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in checkMaxNeighbors(): "+e.Message);
			}
		}

		private void cutOffProperNode(string p_node, ArrayList p_neighbors, Hashtable p_node_list)
		{
			try
			{
				Point node_point = getPointFromNodeList(p_node, p_node_list);
				string ProperNode = "NA";
				int maxNeighbors = 0;
				for(int i=0; i<p_neighbors.Count; i++)
				{
					string nb = p_neighbors[i].ToString();
					int numOfNeighbors = getNeighbors(nb).Count;
					if(maxNeighbors<numOfNeighbors)
					{
						maxNeighbors = numOfNeighbors;
						ProperNode = nb;
					}
				}
				removeProperNode(p_node, ProperNode, p_node_list);   
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in cutOffFarthestNode(): "+e.Message);
			}
		}

		private void removeProperNode(string p_node, string p_properNode, Hashtable p_node_list)
		{
			try
			{
				IDictionaryEnumerator ide = p_node_list.GetEnumerator();
				ArrayList al = new ArrayList();
				int i  = 0;
				while(ide.MoveNext())
				{
					if(ide.Key.ToString().Equals(p_node))
					{
						al = getNeighbors(p_node);
						al.Remove(p_properNode);
						neighbors_hash.Remove(p_node);
						neighbors_hash.Add(p_node, al);
						i++;
					}
					else if(ide.Key.ToString().Equals(p_properNode))
					{
						al = getNeighbors(p_properNode);
						al.Remove(p_node);
						neighbors_hash.Remove(p_properNode);
						neighbors_hash.Add(p_properNode, al);
						i++;
					}
					else if(i==2)
					{
						break;
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in removeFarthestNode(): "+e.Message);
			}

		}

		private Point getPointFromNodeList(string p_node, Hashtable p_node_list)
		{
			IDictionaryEnumerator ide = p_node_list.GetEnumerator();
			Point point = new Point();
			while(ide.MoveNext())
			{
				if(ide.Key.ToString().Equals(p_node))
				{
					point = (Point)ide.Value;
					break;
				}
			}
			return point;
		}

		public void showInsideVisited()
		{
			string inside ="";
			IEnumerator ie = visited.GetEnumerator();
			while(ie.MoveNext())
			{
				inside = inside+ie.Current.ToString()+" ";
			}
			MessageBox.Show("Inside Visited : "+inside);
		}

		public void showInsideNeighbors_hash()
		{
			try
			{
				string inside ="";
				IDictionaryEnumerator ide = neighbors_hash.GetEnumerator();
				while(ide.MoveNext())
				{
					inside = inside+ide.Key.ToString()+": ";
					ArrayList al = new ArrayList();
					al = (ArrayList)ide.Value;
					IEnumerator ie = al.GetEnumerator();
					while(ie.MoveNext())
					{
						inside = inside+ie.Current.ToString()+" ";
					}
					inside = inside+"\r\n";
				}
				MessageBox.Show("Inside Neighbors_hash: "+"\r\n"+inside);
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception in showInsideNeighbors_hash() : "+e.Message);
			}
		}

		private bool isSamePoint(Point p1, Point p2)
		{
			if(((p1.row==p2.row) && (p1.col==p2.col)))
			{
				return true;               
			}
			return false;
		}
		public ArrayList getNeighbors(string p_node)
		{
			IDictionaryEnumerator ide = neighbors_hash.GetEnumerator();
			ArrayList al = new ArrayList();
			while(ide.MoveNext())
			{
				if(ide.Key.ToString().Equals(p_node))
				{
					al = (ArrayList)ide.Value;
					break;
				}
			}
			return al;
		}
		
	}//class of NetworkStatus

}//namespace
