/* Author: Jin-Hee Cho 
 * Project Name: Wireless Ad Hoc Messenger Using Pocket PCs
 * Purpose: Thesis for Master Degree in Computer Science at Virginia Polytechnic and State University
 * Funded by Microsoft
 * Date Updated: 7/14/2004
*/
using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Xml;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.IO;
using System.Text;

namespace UserChat
{
	/// <summary>
	/// Summary description for PocketPCChat.
	/// </summary>
	public class PocketPCChat : System.Windows.Forms.Form
	{
		private ControlInvoker controlInvoker;
		private System.Windows.Forms.Panel ChatPanel;
		private System.Windows.Forms.Button btnExit2;
		private System.Windows.Forms.Button btnHangUp;
		private System.Windows.Forms.TextBox MessageTextBox;
		private System.Windows.Forms.TextBox ChatBox;
		private System.Windows.Forms.Button btnSend;
		private System.Windows.Forms.Label label3;
		private System.Windows.Forms.Panel StartPanel;
		private System.Windows.Forms.Button btnExit;
		private System.Windows.Forms.Button btnChat;
		private System.Windows.Forms.Label lblMyID;
		private System.Windows.Forms.Label lblMyIP;
		private System.Windows.Forms.ListView BuddyList;
		private System.Windows.Forms.MainMenu mainMenu1;
		public System.Windows.Forms.TextBox MyID_IP;
		public System.Windows.Forms.TextBox BuddyID_IP;
		private System.Windows.Forms.Label label2;
		private System.Windows.Forms.Label label4;		
		private System.Windows.Forms.CheckBox Conf_CB;

		//***********************************************************************//
		//UdpServer
		//***********************************************************************//
		private Socket udpSocket;
		private Thread udpListeningThread;
		public string ReceivedData;
		public int ReceivedDataLength;
	
		//********************************************************************//
		//UdpClient
		private UdpClient udpClient;
		private bool udpServerIsAlive;
		//***********************************************************************//
		//Variables to hold in User Interface
		public string strMyID="NA";
		public string strMyIP="NA";
		public string strBuddyID="NA";
		public string strBuddyIP="NA";
		private string strMessageBox="";		
		//Show buddyList in ListView
		private Hashtable hashBuddyList;
		//Route Cache to store identified paths for each destination node
		public RouteCache routeCache;
		
		//XmlDoc accessible from anywhere
		private XmlDocument xmlDoc;

		//One sequence for a certain ID
		//private ArrayList sequence;
		//It contains source ID as a Key and sequence ArrayList as a Value
		private Hashtable hashSequence;
		//ArrayList neighbors
		private ArrayList neighbors;
		//Request ID
		private int sequenceNum = 0;
		public Hashtable send_buffer;
		public Hashtable ReceiptBuffer;
		private Hashtable RRTimerBuffer;
		private Hashtable ReceiptTimerBuffer;
		private Hashtable source_neighbors;
		public Hashtable RRTimer2Buffer;
		public Hashtable ReceiptTimer2Buffer;
		public ArrayList OnCallingBuddy;
		
		private string answer;		
		private bool acceptedPacket = false;		
		private string MySessionID = ""; 
		private string BuddySessionID = "";
		public int processedPacketSN;
		private int discoverSeqNum;		
		public DateTime startingDateTime;
		private bool OnACM;
		private bool OnCache;
		private bool isStartPacket;
		private int DataNum = 1;
		private bool isDoneConnectionEstablishment;
		public ArrayList SenderWindow;
		private ArrayList ReceiverWindow;
		private ArrayList processingRDBuffer;
		private int SWS;
		private bool OnReceiptPacket;
		private System.Windows.Forms.Button Back_Btn;
		private bool isSourceNode = false;
		private bool UseOfConf = false;

		public PocketPCChat()
		{
			//
			// Required for Windows Form Designer support
			//
			InitializeComponent();

			try
			{
				processingRDBuffer = new ArrayList();
				//run UdpServer
				start_server();			
				controlInvoker = new ControlInvoker (this);	
				hashBuddyList = new Hashtable();
				xmlDoc = new XmlDocument();
				xmlDoc.Load(new StreamReader("\\My Documents\\Personal\\conf.xml"));
				hashSequence = new Hashtable();
				storeSequenceInHashtable();
				neighbors = new ArrayList();
				send_buffer = new Hashtable();
				ReceiptBuffer = new Hashtable(); 
				RRTimerBuffer = new Hashtable();
				RRTimer2Buffer = new Hashtable();
				ReceiptTimerBuffer = new Hashtable();
				ReceiptTimer2Buffer = new Hashtable();
				source_neighbors = new Hashtable();
				OnCallingBuddy = new ArrayList();
				processedPacketSN = -9;
				discoverSeqNum = -3;				
				isStartPacket = true;
				isDoneConnectionEstablishment = false;
				SenderWindow = new ArrayList();
				ReceiverWindow = new ArrayList();	
				OnReceiptPacket = false;
			}
			catch(Exception e)
			{
				MessageBox.Show("PocketPCChat() Exception is occurred: "+e.Message);
				stop_server();
			}
		}
		/// <summary>
		/// When the frame is loaded, this method is implemented.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void PocketPCChat_Load(object sender, System.EventArgs e)
		{
			try
			{
				//My IP Address
				IPHostEntry localHostEntry = Dns.GetHostByName(Dns.GetHostName());
				IPAddress myIP = localHostEntry.AddressList[0];
				strMyIP = myIP.ToString();
				lblMyIP.Text = "My IP: "+strMyIP;
				//My ID
				strMyID = FindIDbyIPAddressInConfFile(strMyIP);
				lblMyID.Text = "My ID: "+strMyID;
				//Set the source in routeCache
				routeCache = new RouteCache();
				//Set hashBuddyList
				setHashBuddyList();
				showBuddyList();	
			
				OnReceiptPacket = false;
				SWS = UDPCONSTANTS.SWS;
				OnCache = true;
				OnACM = true;

			}
			catch(Exception se)
			{
				MessageBox.Show("PocketPCChat_Load() Exception occurred : "+se.Message);
			}
		}

		/// <summary>
		/// Set hashBuddyList(Hashtable), id as key and ip as value
		/// </summary>
		public void setHashBuddyList()
		{
			try
			{
				XmlNode root = xmlDoc.DocumentElement;
				XmlNodeList xNodes = root.ChildNodes;
				foreach(XmlNode childNode in xNodes)
				{					
					XmlElement currentElement = (XmlElement)childNode;
					string id = currentElement.GetAttribute("ID").ToString();
					string ip = currentElement.GetAttribute("IP").ToString();
					hashBuddyList.Add(id,ip);
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in setHashBuddyList() : "+e.Message);
			}
		}
		public Hashtable getHashBuddyList()
		{
			return hashBuddyList;
		}
		
		
		/// <summary>
		/// Get Pocket PC ID by supplying its IP address
		/// </summary>
		/// <param name="strNodeIPAddress">Pocket PC IP Address in x.x.x.x format</param>
		/// <returns>ID matching input IP address</returns>
		public string FindIDbyIPAddressInConfFile(string strNodeIPAddress)
		{			
			string strReturnID="NA";
			try
			{
				XmlNode root = xmlDoc.DocumentElement;//configuration root element
				XmlNodeList xNodes = root.ChildNodes;
											
				foreach(XmlNode childNode in xNodes)
				{
					
					XmlElement currentElement = (XmlElement)childNode;
					if(currentElement.GetAttribute("IP").Equals(strNodeIPAddress))
					{
						strReturnID=currentElement.GetAttribute("ID").ToString();
					}
				}									
			}
			catch(Exception ex)
			{
				MessageBox.Show("FindIDbyIPAddress: An Exception has occured. \r\n" + ex.ToString()+ex.Message.ToString(), 
					"Pocket Chat",MessageBoxButtons.OK , MessageBoxIcon.Hand ,MessageBoxDefaultButton.Button1);
			}
			return strReturnID;
		}

		/// <summary>
		/// Get IP address with ID input
		/// </summary>
		/// <param name="strID"></param>
		/// <returns>IP address matching with ID input</returns>
		public string GetIPAddressByIDInConfFile(string strID)
		{
			string strReturnIP="NA";
			try
			{
				XmlNode root = xmlDoc.DocumentElement;//configuration root element
				XmlNodeList xNodes = root.ChildNodes;
											
				foreach(XmlNode childNode in xNodes)
				{
					
					XmlElement currentElement = (XmlElement)childNode;
					if(currentElement.GetAttribute("ID").Equals(strID))
					{
						strReturnIP=currentElement.GetAttribute("IP").ToString();
					}
				}									
			}
			catch(Exception ex)
			{
				MessageBox.Show("FindIDbyIPAddress: An Exception has occured. \r\n" + ex.ToString()+ex.Message.ToString(), 
					"Pocket Chat",MessageBoxButtons.OK , MessageBoxIcon.Hand ,MessageBoxDefaultButton.Button1);
			}
			return strReturnIP;
		}

		
		public ArrayList getNeighbors(string p_ID)
		{
			ArrayList neighbor = new ArrayList();
			Hashtable node_list = new Hashtable();
			
			try
			{
				
				if(UseOfConf)
				{
					NetworkStatus ns = new NetworkStatus(hashSequence);
					IDictionaryEnumerator ide = hashSequence.GetEnumerator();
				
					int T = 0;
					if(isStartPacket)
					{
						T = 2;//While connection is established, this specific NT will be used.	
					}
					else
					{
						T= getNTTime();
					}
					while(ide.MoveNext())
					{
						string node_id = ide.Key.ToString();
						Point point = new Point();
						point = ns.getLocation(T, node_id);
						node_list.Add(node_id, point);
					}//while
					//ns.showNodeList(node_list);
					ns.adjustNetworkTopology(node_list, OnACM, isStartPacket);
					//MessageBox.Show("After adjusting NT");
					//ns.showInsideNeighbors_hash();	
					neighbor = ns.getNeighbors(strMyID);
				}
				else
				{
					IDictionaryEnumerator ide = hashBuddyList.GetEnumerator();
					for(int i = 0; ide.MoveNext(); i++)
					{
						string node = ide.Key.ToString();
						if(!node.Equals(strMyID))
						{
							neighbor.Add(node);
						}
					}
				}
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception in getNeighbors() PocketPCChat : "+e.Message);
			}
			return neighbor;
		}//getNeightbors()

		//*************************************************************************/
		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			base.Dispose( disposing );
		}

		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.ChatPanel = new System.Windows.Forms.Panel();
			this.Back_Btn = new System.Windows.Forms.Button();
			this.label4 = new System.Windows.Forms.Label();
			this.label2 = new System.Windows.Forms.Label();
			this.BuddyID_IP = new System.Windows.Forms.TextBox();
			this.MyID_IP = new System.Windows.Forms.TextBox();
			this.btnExit2 = new System.Windows.Forms.Button();
			this.btnHangUp = new System.Windows.Forms.Button();
			this.ChatBox = new System.Windows.Forms.TextBox();
			this.MessageTextBox = new System.Windows.Forms.TextBox();
			this.btnSend = new System.Windows.Forms.Button();
			this.label3 = new System.Windows.Forms.Label();
			this.StartPanel = new System.Windows.Forms.Panel();
			this.Conf_CB = new System.Windows.Forms.CheckBox();
			this.BuddyList = new System.Windows.Forms.ListView();
			this.btnExit = new System.Windows.Forms.Button();
			this.btnChat = new System.Windows.Forms.Button();
			this.lblMyID = new System.Windows.Forms.Label();
			this.lblMyIP = new System.Windows.Forms.Label();
			this.mainMenu1 = new System.Windows.Forms.MainMenu();
			// 
			// ChatPanel
			// 
			this.ChatPanel.BackColor = System.Drawing.Color.FromArgb(((System.Byte)(0)), ((System.Byte)(192)), ((System.Byte)(0)));
			this.ChatPanel.Controls.Add(this.Back_Btn);
			this.ChatPanel.Controls.Add(this.label4);
			this.ChatPanel.Controls.Add(this.label2);
			this.ChatPanel.Controls.Add(this.BuddyID_IP);
			this.ChatPanel.Controls.Add(this.MyID_IP);
			this.ChatPanel.Controls.Add(this.btnExit2);
			this.ChatPanel.Controls.Add(this.btnHangUp);
			this.ChatPanel.Controls.Add(this.ChatBox);
			this.ChatPanel.Controls.Add(this.MessageTextBox);
			this.ChatPanel.Controls.Add(this.btnSend);
			this.ChatPanel.Controls.Add(this.label3);
			this.ChatPanel.Location = new System.Drawing.Point(0, 296);
			this.ChatPanel.Size = new System.Drawing.Size(240, 272);
			// 
			// Back_Btn
			// 
			this.Back_Btn.Location = new System.Drawing.Point(160, 240);
			this.Back_Btn.Size = new System.Drawing.Size(72, 24);
			this.Back_Btn.Text = "Back";
			this.Back_Btn.Click += new System.EventHandler(this.Start_Btn_Click);
			// 
			// label4
			// 
			this.label4.Font = new System.Drawing.Font("Tahoma", 9F, System.Drawing.FontStyle.Bold);
			this.label4.Location = new System.Drawing.Point(8, 32);
			this.label4.Size = new System.Drawing.Size(56, 24);
			this.label4.Text = "Buddy";
			// 
			// label2
			// 
			this.label2.Font = new System.Drawing.Font("Tahoma", 9F, System.Drawing.FontStyle.Bold);
			this.label2.Location = new System.Drawing.Point(8, 8);
			this.label2.Size = new System.Drawing.Size(56, 24);
			this.label2.Text = "Myself";
			// 
			// BuddyID_IP
			// 
			this.BuddyID_IP.Location = new System.Drawing.Point(64, 32);
			this.BuddyID_IP.ReadOnly = true;
			this.BuddyID_IP.Size = new System.Drawing.Size(160, 22);
			this.BuddyID_IP.Text = "";
			// 
			// MyID_IP
			// 
			this.MyID_IP.Location = new System.Drawing.Point(64, 8);
			this.MyID_IP.ReadOnly = true;
			this.MyID_IP.Size = new System.Drawing.Size(160, 22);
			this.MyID_IP.Text = "";
			// 
			// btnExit2
			// 
			this.btnExit2.Location = new System.Drawing.Point(88, 240);
			this.btnExit2.Size = new System.Drawing.Size(64, 24);
			this.btnExit2.Text = "EXIT";
			this.btnExit2.Click += new System.EventHandler(this.btnExit2_Click);
			// 
			// btnHangUp
			// 
			this.btnHangUp.Location = new System.Drawing.Point(8, 240);
			this.btnHangUp.Size = new System.Drawing.Size(72, 24);
			this.btnHangUp.Text = "HANG UP";
			this.btnHangUp.Click += new System.EventHandler(this.btnHangUp_Click);
			// 
			// ChatBox
			// 
			this.ChatBox.ForeColor = System.Drawing.Color.Transparent;
			this.ChatBox.Location = new System.Drawing.Point(8, 128);
			this.ChatBox.Multiline = true;
			this.ChatBox.ReadOnly = true;
			this.ChatBox.ScrollBars = System.Windows.Forms.ScrollBars.Vertical;
			this.ChatBox.Size = new System.Drawing.Size(224, 104);
			this.ChatBox.Text = "";
			// 
			// MessageTextBox
			// 
			this.MessageTextBox.Location = new System.Drawing.Point(8, 88);
			this.MessageTextBox.Multiline = true;
			this.MessageTextBox.ScrollBars = System.Windows.Forms.ScrollBars.Vertical;
			this.MessageTextBox.Size = new System.Drawing.Size(224, 32);
			this.MessageTextBox.Text = "Hello1";
			// 
			// btnSend
			// 
			this.btnSend.Location = new System.Drawing.Point(160, 56);
			this.btnSend.Size = new System.Drawing.Size(72, 24);
			this.btnSend.Text = "SEND";
			this.btnSend.Click += new System.EventHandler(this.btnSend_Click);
			// 
			// label3
			// 
			this.label3.Font = new System.Drawing.Font("Microsoft Sans Serif", 9F, System.Drawing.FontStyle.Bold);
			this.label3.Location = new System.Drawing.Point(8, 56);
			this.label3.Size = new System.Drawing.Size(128, 24);
			this.label3.Text = "Type  Your  Message";
			// 
			// StartPanel
			// 
			this.StartPanel.BackColor = System.Drawing.Color.Gold;
			this.StartPanel.Controls.Add(this.Conf_CB);
			this.StartPanel.Controls.Add(this.BuddyList);
			this.StartPanel.Controls.Add(this.btnExit);
			this.StartPanel.Controls.Add(this.btnChat);
			this.StartPanel.Controls.Add(this.lblMyID);
			this.StartPanel.Controls.Add(this.lblMyIP);
			this.StartPanel.Size = new System.Drawing.Size(240, 296);
			// 
			// Conf_CB
			// 
			this.Conf_CB.Checked = true;
			this.Conf_CB.CheckState = System.Windows.Forms.CheckState.Checked;
			this.Conf_CB.Location = new System.Drawing.Point(8, 200);
			this.Conf_CB.Size = new System.Drawing.Size(216, 20);
			this.Conf_CB.Text = "Use of Configuration file";
			this.Conf_CB.CheckStateChanged += new System.EventHandler(this.Conf_CB_CheckStateChanged);
			// 
			// BuddyList
			// 
			this.BuddyList.Location = new System.Drawing.Point(8, 32);
			this.BuddyList.Size = new System.Drawing.Size(224, 160);
			// 
			// btnExit
			// 
			this.btnExit.Location = new System.Drawing.Point(128, 224);
			this.btnExit.Size = new System.Drawing.Size(96, 24);
			this.btnExit.Text = "EXIT";
			this.btnExit.Click += new System.EventHandler(this.btnExit_Click);
			// 
			// btnChat
			// 
			this.btnChat.Location = new System.Drawing.Point(8, 224);
			this.btnChat.Size = new System.Drawing.Size(112, 24);
			this.btnChat.Text = "CHAT";
			this.btnChat.Click += new System.EventHandler(this.btnChat_Click);
			// 
			// lblMyID
			// 
			this.lblMyID.Font = new System.Drawing.Font("Microsoft Sans Serif", 8.25F, System.Drawing.FontStyle.Bold);
			this.lblMyID.Location = new System.Drawing.Point(8, 8);
			this.lblMyID.Size = new System.Drawing.Size(88, 24);
			this.lblMyID.Text = "My ID :";
			// 
			// lblMyIP
			// 
			this.lblMyIP.Font = new System.Drawing.Font("Microsoft Sans Serif", 8.25F, System.Drawing.FontStyle.Bold);
			this.lblMyIP.Location = new System.Drawing.Point(96, 8);
			this.lblMyIP.Size = new System.Drawing.Size(136, 24);
			this.lblMyIP.Text = "My IP :";
			// 
			// PocketPCChat
			// 
			this.ClientSize = new System.Drawing.Size(240, 568);
			this.Controls.Add(this.ChatPanel);
			this.Controls.Add(this.StartPanel);
			this.Menu = this.mainMenu1;
			this.Text = "PocketPCChat";
			this.Load += new System.EventHandler(this.PocketPCChat_Load);

		}
		#endregion


		private void btnExit2_Click(object sender, System.EventArgs e)
		{
			exit();
		}

		private void resetAll()
		{
			isSourceNode = false;
			isDoneConnectionEstablishment = false;
			SenderWindow.Clear();
			ReceiverWindow.Clear();
			BuddyID_IP.Text = "";
			MyID_IP.Text = "";
            MessageTextBox.Text = "Hello1";			
			acceptedPacket = false;
			sequenceNum = 0;
			send_buffer.Clear();
			ReceiptBuffer.Clear();
			ChatBox.Text = "";
			MySessionID = "";
			BuddySessionID = "";
			ReceiptTimerBuffer.Clear();
			processedPacketSN = -9;
			discoverSeqNum = -3;
			OnCallingBuddy.Clear();
			ReceiptTimer2Buffer.Clear();
			RRTimer2Buffer.Clear();
			routeCache.Clear();			
		}

		private void btnHangUp_Click(object sender, System.EventArgs e)
		{
			//if(MessageTextBox.Text.ToString().Length==0)
			//{
				if(isDoneConnectionEstablishment)
				{
					Packet packet = new Packet();
					SendHangUpPacket(ref packet);
				}
				//Back to the Start Panel
				showBuddyScreen();
				//resetAll();
			//}
			//else
			//{
			//	MessageBox.Show("You need to send a message first ! ");
			//}
		}

		/// <summary>
		/// When the Send Button is clicked, the packet with a message is sent.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void btnSend_Click(object sender, System.EventArgs e)
		{
			try
			{
				//MessageBox.Show("Buddy: "+BuddyID_IP.Text+"\r\nMyself: "+MyID_IP.Text+"\r\n"+ChatInfoShowTimer.Enabled);

				if(MessageTextBox.Text.ToString().Length>0) 
				{
					if(SenderWindow.Count<SWS)
					{
						strMessageBox = MessageTextBox.Text;
						DataNum++;
						Packet packet = new Packet();
						neighbors = getNeighbors(strMyID);

						if(!routeCache.isEmpty())
						{
							SendMessagePacket(ref packet);
						}
						else if(neighbors.Count==0)
						{
							MessageBox.Show("Can't Find Any Route to "+packet.strTarget, "No Route Found",
								MessageBoxButtons.OK, MessageBoxIcon.Exclamation, MessageBoxDefaultButton.Button1);
						}
						else
						{
							packet.strType = Packet_CONSTANTS.DATA_PACKET;
							packet.strTarget = strBuddyID;
							packet.strInitiator = strMyID;
							packet.strRequestID = getRequestID(Packet_CONSTANTS.DATA_PACKET);
							packet.strRouteCache = Packet_CONSTANTS.NOTAVAILABLE;
							packet.strRouteRecord = strMyID;
							packet.strPayload = strMessageBox;
							packet.strCurrentID = strMyID;
							SaveInSendBuffer(packet);							
							PerformRouteDiscovery(packet);
						}
						MessageTextBox.Text = "Hello"+DataNum;				
					}
					else
					{
                        //MessageBox.Show("Please slow down to send your message!");
						ChatBox.Text += "Please slow down to send your message! \r\n";						
					}
				}
				else
				{
                    MessageBox.Show("Please enter a message first!");  
				}
			}
			catch(Exception ex)
			{
				MessageBox.Show("Exception in SendButton():"+ex.Message);
			}
			
		}


		/// <summary>
		/// Save the packet to send in send_buffer unless the send_buffer
		/// already has the packet and its request id or crrent node is not initiator.
		/// </summary>
		/// <param name="packet">packet passed</param>
		private void SaveInSendBuffer(Packet packet)
		{
			try
			{
				if(!send_buffer.ContainsKey(packet.strRequestID))
				{
					send_buffer.Add(packet.strRequestID, packet);
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SaveInSendBuffer() :"+e.Message);
			}
		}

		private void SetIsStartPacket(Packet p_packet)
		{
			if(!p_packet.strType.Equals(Packet_CONSTANTS.START_CHAT_PACKET) &&
				!p_packet.strType.Equals(Packet_CONSTANTS.REJECT_START_CHAT_PACKET) &&
				!p_packet.strType.Equals(Packet_CONSTANTS.ACCEPT_START_CHAT_PACKET))
			{
				isStartPacket = false;	
			}
		}
		
		/// <summary>
		/// In Send Button, Packet with a message is sent to target node(BuddyID)
		/// </summary>
		/// <param name="packet">Packet with a message</param>
		public void SendPacket(ref Packet packet)
		{
			try
			{
				//SetACM();
				SetIsStartPacket(packet);				
				XmlDocument doc = new XmlDocument();
				doc.LoadXml("<Packet></Packet>");

				XmlNode root = doc.DocumentElement;

				//1. Type
				XmlElement elem = doc.CreateElement(Packet_CONSTANTS.TYPE);
				elem.InnerText = packet.strType;
				root.AppendChild(elem);


				//2.Initiator
				elem = doc.CreateElement(Packet_CONSTANTS.INITIATOR);
				elem.InnerText = packet.strInitiator;
				root.AppendChild(elem);


				//3.Current ID
				elem = doc.CreateElement(Packet_CONSTANTS.CURRENTID);
				elem.InnerText = packet.strCurrentID;
				root.AppendChild(elem);
				
				//4.Payload
				elem = doc.CreateElement(Packet_CONSTANTS.PAYLOAD);
				elem.InnerText = packet.strPayload;
				root.AppendChild(elem);

				//5. RequestID
				elem = doc.CreateElement(Packet_CONSTANTS.REQUESTID);
				elem.InnerText = packet.strRequestID;
				root.AppendChild(elem);

				//6.RouteCache
				elem = doc.CreateElement(Packet_CONSTANTS.ROUTECACHE);
				elem.InnerText = packet.strRouteCache;
				root.AppendChild(elem);

				//7.RouteRecord
				elem = doc.CreateElement(Packet_CONSTANTS.ROUTERECORD);
				elem.InnerText = packet.strRouteRecord;
				root.AppendChild(elem);

				//8.Target
				elem = doc.CreateElement(Packet_CONSTANTS.TARGET);
				elem.InnerText = packet.strTarget;
				root.AppendChild(elem);
				
				//9.ReceiptID
				elem = doc.CreateElement(Packet_CONSTANTS.RECEIPTID);
				elem.InnerText = packet.strReceiptID;
				root.AppendChild(elem);

				//for debugging
				string messageToSend = "";
				string serverID = "";
				string serverIP = "";
				bool sentSuccess = false;
				neighbors = getNeighbors(strMyID);					

				if(packet.strRouteCache.Equals(Packet_CONSTANTS.NOTAVAILABLE))
				{//when there is no available path in route cache, broadcast to 
				 //all neighboring nodes

					if(neighbors.Count>0)
					{
						for(int i=0; i<neighbors.Count; i++)
						{
							serverID = neighbors[i].ToString();
							if(isProperNextNodeInRD(serverID, packet.strRouteRecord))
							{
								//MessageBox.Show(" packet type and serverID: "+packet.strType+" "+serverID+" i "+i);
								//9.Replace ReceiptID
								packet.strReceiptID = getReceiptID(strMyID, serverID);
								elem.InnerText = packet.strReceiptID;
								root.AppendChild(elem);

								//10. Put TimeStamp into packet
								elem = doc.CreateElement(Packet_CONSTANTS.TIMESTAMP);
								if(packet.strInitiator.Equals(strMyID))
								{
									//packet.strTimeStamp = NTTimer.getCount().ToString();
									packet.strTimeStamp = getNTTime().ToString();
								}
								elem.InnerText = packet.strTimeStamp;
								root.AppendChild(elem);

								SaveInReceiptBuffer(packet);//for a receipt packet

								serverIP = GetIPAddressByIDInConfFile(serverID);
								messageToSend = doc.InnerXml.ToString();
								//showDebugSendingPackets(packet);
								sentSuccess = sendUDP(serverIP, messageToSend);
								/*if(sentSuccess == false)//Route Error Occurs.
								{
									ReleaseReceiptTimer(packet.strReceiptID);
									InformRouteErrorToSource(packet);
								}*/
							}//Proper Next Node if
						}//end of for
					}
					else
					//neighbors.Count == 0: when there is no neighbors, 
					//send route error packet to the source
					{
						serverID = "NA";
						//9.Replace ReceiptID
						packet.strReceiptID = getReceiptID(strMyID, serverID);
						elem.InnerText = packet.strReceiptID;
						root.AppendChild(elem);

						//10. Put TimeStamp into packet
						elem = doc.CreateElement(Packet_CONSTANTS.TIMESTAMP);
						if(packet.strInitiator.Equals(strMyID))
						{
							//packet.strTimeStamp = NTTimer.getCount().ToString();
							packet.strTimeStamp = getNTTime().ToString();
						}
						elem.InnerText = packet.strTimeStamp;
						root.AppendChild(elem);

						SaveInReceiptBuffer(packet);//for a receipt packet
						ReleaseReceiptTimer(packet.strReceiptID);
						InformRouteErrorToSource(packet);
					}
				}
				else
				{//When route cache has a path to reach target
									
					serverID = getNextNodeInCachePath(packet.strRouteCache);
					serverIP = GetIPAddressByIDInConfFile(serverID);

					//9.Replace ReceiptID
					packet.strReceiptID = getReceiptID(strMyID, serverID);
					elem.InnerText = packet.strReceiptID;
					root.AppendChild(elem);

					//10. Put TimeStamp into packet
					elem = doc.CreateElement(Packet_CONSTANTS.TIMESTAMP);
					if(packet.strInitiator.Equals(strMyID))
					{
						//packet.strTimeStamp = NTTimer.getCount().ToString();
						packet.strTimeStamp = getNTTime().ToString();
					}
					elem.InnerText = packet.strTimeStamp;
					root.AppendChild(elem);

					SaveInReceiptBuffer(packet);//for a receipt packet
					if(neighbors.Contains(serverID))
					{
						messageToSend = doc.InnerXml.ToString();
						//showDebugSendingPackets(packet);
						sentSuccess = sendUDP(serverIP, messageToSend);
						/*if(sentSuccess==false)
						{
							ReleaseReceiptTimer(packet.strReceiptID);
							InformRouteErrorToSource(packet);
						}*/
					}
					else
					{
						ReleaseReceiptTimer(packet.strReceiptID);
						InformRouteErrorToSource(packet);
					}
				}
			}//try
			catch(Exception e)
			{
				MessageBox.Show("Exception is occurred in SendPacket() : "+e.Message);
			}

		}//end of SendPacket()
	    
		private void showDebugSendingPackets(Packet p_packet)
		{
			try
			{
				if(!p_packet.strType.Equals(Packet_CONSTANTS.START_CHAT_PACKET) &&
				   !p_packet.strType.Equals(Packet_CONSTANTS.REJECT_START_CHAT_PACKET) &&
				   !p_packet.strType.Equals(Packet_CONSTANTS.ACCEPT_START_CHAT_PACKET))
				{
					ChatBox.Text = ChatBox.Text+"\r\n"+routeCache.showRouteCacheInside();
					ChatBox.Text = ChatBox.Text+"\r\n SEND: "+ p_packet.strInitiator+": "+
						p_packet.strTimeStamp+" "+
						p_packet.strType+" "+p_packet.strRequestID+
						" "+p_packet.strRouteCache; 
				}
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in showDebugSendingPackets(): "+e.Message); 
			}
		}

		private void showInsideNeighbor()
		{
            IEnumerator ie = neighbors.GetEnumerator();
			string inside = "";
			while(ie.MoveNext())
			{
               inside = inside + ie.Current.ToString()+" ";
			}
			MessageBox.Show("Inside Neighbor in "+strMyID+" : "+inside);
		}

		private bool isProperNextNodeInRD(string p_serverID, string p_RouteRecord)
		{
			try
			{
				for(int i=0; i<p_RouteRecord.Length; i++)
				{
					if(p_serverID.Equals(p_RouteRecord[i].ToString()))
					{
						return false;
					}
				}
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception in isProperNextNodeInRD(): "+e.Message);
			}
			return true;
		}
		
		
		private void SetMySessionID()
		{			
			MySessionID = getSessionID()+":"+strMyID+":"+strBuddyID;
		}

		private void btnChat_Click(object sender, System.EventArgs e)
		{
			try
			{
				SWS = UDPCONSTANTS.SWS;
				strBuddyID = BuddyList.Items[(BuddyList.SelectedIndices[0])].SubItems[0].Text;				
				strBuddyIP = BuddyList.Items[(BuddyList.SelectedIndices[0])].SubItems[1].Text;
				
				if(!strBuddyID.Equals("N/A"))
				{
					SendBroadcastNTPacket();
					isSourceNode = true;
				}
				else
				{
					MessageBox.Show("Please select a buddy to chat!");
				}
			}
			catch(Exception ex)
			{
                MessageBox.Show("Exception in btnChat_Click(): "+ex.Message);
			}
		}

		private void SendBroadcastNTPacket()
		{
			try
			{

				BroadcastPacket broadPacket = new BroadcastPacket();
				broadPacket.strType = Packet_CONSTANTS.BROADCAST_NT_PACKET;				
				//broadPacket.strOnACM = ACM_CB.Checked.ToString();
				//broadPacket.strOnCache = Cache_CB.Checked.ToString();
				//broadPacket.strSWS = UDPCONSTANTS.SWS.ToString();
				//broadPacket.strOnRP = RP_CB.Checked.ToString();		
				broadPacket.strOnConf = Conf_CB.Checked.ToString();
				
				string broadcastNT = CreateBroadcastPacket(broadPacket);
				IDictionaryEnumerator ide = hashBuddyList.GetEnumerator();
				while(ide.MoveNext())
				{
					string serverID = ide.Key.ToString();					
					string serverIP = GetIPAddressByIDInConfFile(serverID);
					sendUDP(serverIP, broadcastNT);
				}	
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in BroadcastNTPacket() : "+e.Message);
			}
		}

		private string CreateBroadcastPacket(BroadcastPacket p_packet)
		{
			string messageToSend ="";
			try
			{
				XmlDocument doc = new XmlDocument();
				doc.LoadXml("<Packet></Packet>");

				XmlNode root = doc.DocumentElement;

				//1. Type
				XmlElement elem = doc.CreateElement(Packet_CONSTANTS.TYPE);
				elem.InnerText = p_packet.strType;
				root.AppendChild(elem);
				//2. Cache on or off
				/*elem = doc.CreateElement(Packet_CONSTANTS.CACHE);
				elem.InnerText = p_packet.strOnCache;
				root.AppendChild(elem);				
				//3. ACM on or off
				elem = doc.CreateElement(Packet_CONSTANTS.ACM);
				elem.InnerText = p_packet.strOnACM;
				root.AppendChild(elem);				
				//4. SWS
				elem = doc.CreateElement(Packet_CONSTANTS.SWS);
				elem.InnerText = p_packet.strSWS;
				root.AppendChild(elem);					
				//5. Use of Receipt Packet
				elem = doc.CreateElement(Packet_CONSTANTS.RP);
				elem.InnerText = p_packet.strOnRP;
				root.AppendChild(elem);		*/
				//6. Use of Configuration File
				elem = doc.CreateElement(Packet_CONSTANTS.CONF);
				elem.InnerText = p_packet.strOnConf;
				root.AppendChild(elem);		

				messageToSend = doc.InnerXml.ToString();				
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in CreateBroadcastPacket(): "+e.Message);
			}
			return messageToSend;

		}

		private void btnExit_Click(object sender, System.EventArgs e)
		{
			exit();
		}

		/// <summary>
		/// When EXIT buttons are pressed, this method is called.
		/// </summary>
		private void exit()
		{
			//udpClient.Close();
			udpServerIsAlive = false;
			stop_server();			
			resetAll();
			this.Close();
		}
  


		/// Return true if the request id is already received. Otherwise, return false.
		/// </summary>
		/// <param name="p_requestID">request ID of the received packet</param>
		/// <returns>boolean</returns>
		private bool isAlreadyReceived(Packet packet)
		{
			try
			{
				//for all kinds of route reply packets: In Sender
				if(packet.strType.Equals(Packet_CONSTANTS.ROUTE_REPLY_PACKET) ||
					packet.strType.Equals(Packet_CONSTANTS.ACCEPT_START_CHAT_PACKET) ||
					packet.strType.Equals(Packet_CONSTANTS.REJECT_START_CHAT_PACKET) ||
					packet.strType.Equals(Packet_CONSTANTS.REPLY_HANG_UP_PACKET) ||
					packet.strType.Equals(Packet_CONSTANTS.REPLY_ROUTE_DISCOVERY_PACKET))
				{
					if(send_buffer.ContainsKey(packet.strRequestID))
					{
						return false;
					}
					return true; 
				}
					//packets from buddy side: In Receiver
				else if(packet.strType.Equals(Packet_CONSTANTS.DATA_PACKET))
				{
					int receivedSN = getSequenceNumFromRequestIDReceived(packet.strRequestID);
					if(receivedSN<=processedPacketSN)
					{
						return true;
					}
					else
					{
						return isInReceiverWindow(receivedSN.ToString());
					}					
				}
				else if(packet.strType.Equals(Packet_CONSTANTS.ROUTE_ERROR_PACKET) ||
					packet.strType.Equals(Packet_CONSTANTS.ROUTE_DISCOVERY_PACKET))
				{
					return false;
				}
				else if(packet.strType.Equals(Packet_CONSTANTS.START_CHAT_PACKET))
				{
					if(isDoneConnectionEstablishment)
					{
						return true;
					}
					return false;
				}
				else if(packet.strType.Equals(Packet_CONSTANTS.HANG_UP_PACKET))
				{
					if(!isDoneConnectionEstablishment)
					{
						return true;
					}
					return false;
				}
				else if(packet.strType.Equals(Packet_CONSTANTS.ONCALLING_PACKET))
				{
					if(send_buffer.Contains(packet.strRequestID))
					{
						return false;
					}
					else
					{
						return true;
					}
				}
				return false;
			}//try			
			catch(Exception e)
			{
				MessageBox.Show("Exception occurred in isAlreadyReceived(): "+e.Message);
				return false;
			}
		}


		private bool isInReceiverWindow(string p_SN)
		{
			bool isHere = false;
			try
			{
				IEnumerator ie = ReceiverWindow.GetEnumerator();
				while(ie.MoveNext())
				{
					ElementsForReceiverWindow ERW = (ElementsForReceiverWindow)ie.Current;
					if(ERW.SN.Equals(p_SN))
					{
						isHere = true;
						break;
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in isInReceiverWindow(): "+e.Message);
			}
			return isHere;
		}

		private void showChatPanel()
		{
			ChatPanel.Top = 0;
			ChatPanel.Visible = true;
			ChatPanel.Focus();
			ChatPanel.BringToFront();
		}

		private void SetReceiptTimer(string p_receiptID, Packet p_packet)
		{
			/*try
			{
				MyTimer myTimer = new MyTimer(p_receiptID, 500, this,p_packet);
				myTimer.setReceiptACKTimer();
				ReceiptTimerBuffer.Add(p_receiptID,myTimer);
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SetReceiptTimer(): "+e.Message);
			}*/
			try
			{
                MyTimer2 myTimer = new MyTimer2(1000, true,1,this,p_packet, p_receiptID);
				myTimer.SetReceiptTimer2();
				ReceiptTimer2Buffer.Add(p_receiptID, myTimer);
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception i SetReceiptTimer(); "+e.Message);
			}
		}

		private void SaveInReceiptBuffer(Packet packet)
		{
			try
			{
				if(OnReceiptPacket)
				{	
					if(packet.strType.Equals(Packet_CONSTANTS.DATA_PACKET))
					{				
						SetReceiptTimer(packet.strReceiptID, packet);
						ReceiptBuffer.Add(packet.strReceiptID, packet);
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SaveInReceiptBuffer(): "+e.Message);
			}
		}

		public void UpdateReceiptBuffer(string p_receiptID)
		{
			try
			{
				if(ReceiptBuffer.ContainsKey(p_receiptID))
				{
					ReceiptBuffer.Remove(p_receiptID);
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception occurred in UpdateReceiptBuffer():"+e.Message);
			}
		}

		/// <summary>
		/// Release Receipt Timer2
		/// </summary>
		/// <param name="p_ID">receiptID of packet</param>
		public void ReleaseReceiptTimer(string p_ID)
		{
			try
			{
				IDictionaryEnumerator ide = ReceiptTimer2Buffer.GetEnumerator();
				while(ide.MoveNext())
				{
					if(ide.Key.ToString().Equals(p_ID))
					{
						MyTimer2 tempTimer = (MyTimer2)ide.Value;	
						tempTimer.setEnabled(false);
						break;
					}
				}
				ReceiptTimer2Buffer.Remove(p_ID);
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception ocurred in ReleaseReceiptTimer():"+e.Message);
			}
		}

		private string showReceiptBufferInside()
		{
			string receiptBufferInside="";
			IDictionaryEnumerator ide = ReceiptBuffer.GetEnumerator();
		    
			while(ide.MoveNext())
			{
				receiptBufferInside = receiptBufferInside+"\r\n"+ide.Key.ToString();

			}
			if(receiptBufferInside.Equals(""))
			{
				receiptBufferInside = "NA";
			}
			return receiptBufferInside;
		}

		private void showBuddyScreen()
		{
			StartPanel.Top = 0;
			StartPanel.Focus();
			StartPanel.Visible = true;
			StartPanel.BringToFront();
		}

		private bool isReceiptPacket()
		{
			string startStringInReceiptPacket = "<Packet><Type>"+Packet_CONSTANTS.RECEIPT_PACKET;
			if(ReceivedDataLength>0)
			{
				if(ReceivedData.StartsWith(startStringInReceiptPacket))
				{
					return true;
				}
			}
			return false;
		}

		private ReceiptPacket transformXmlIntoReceiptPacket()
		{
			ReceiptPacket rcv_packet = new ReceiptPacket();
			try
			{
				//Monitor.Enter(this);
				XmlDocument doc = new XmlDocument();
				//P();//Semaphore P
				doc.LoadXml(ReceivedData);
				ReceivedDataLength=0;//reset the received data's length
				//V();//Semaphore V
				XmlNode root = doc.DocumentElement; 
				XmlNodeList xNodes = root.ChildNodes;
			
				foreach(XmlNode childNode in xNodes)
				{
					
					XmlElement elem = (XmlElement)childNode;
					if(elem.Name.Equals(Packet_CONSTANTS.TYPE))
					{
						rcv_packet.strType = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.RECEIPTID))
					{
						rcv_packet.strReceiptID = elem.FirstChild.Value;
					}
				}	
				////Monitor.Exit(this);
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception is occurred in transformXmlIntoPacket() : "+e.Message);
			}
			return rcv_packet;			
		}
		

		private bool isBroadcastNTPacket()
		{
			try
			{
				string startString = "<Packet><Type>"+Packet_CONSTANTS.BROADCAST_NT_PACKET;
			
				if(ReceivedData.StartsWith(startString))
				{
					return true; 
				}
				else
				{
					return false; 
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in isBroadcastNTPacket(): "+e.Message);
				return false;
			}
		}

		public int getNTTime()
		{
            long timeElapsed = (DateTime.Now.Ticks - startingDateTime.Ticks)/10000000;
			string strTimeElapsed = timeElapsed.ToString();
			int intTimeElapsed = int.Parse(strTimeElapsed);
			return intTimeElapsed;
		}

		private void startTickingNTTime()
		{
			startingDateTime = DateTime.Now;
		}

		private void SendOnCallingPacket(ref Packet packet)
		{
			try
			{
				string initiator = packet.strTarget;
				string target = packet.strInitiator;

				packet.strInitiator = initiator;
				packet.strTarget = target;

				if(!routeCache.isEmpty())
				{
					packet.strRouteCache = routeCache.getPath();
				}
				else
				{
					packet.strRouteCache = Packet_CONSTANTS.NOTAVAILABLE;
				}
				packet.strRouteRecord = strMyID;
				packet.strCurrentID = strMyID;
				packet.strPayload = Packet_CONSTANTS.NOTAVAILABLE;
				packet.strType = Packet_CONSTANTS.ONCALLING_PACKET;
				SendPacket(ref packet);						    
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in SendOnCallingPacket(): "+e.Message);    
			}
		}
		


		private bool isProperBuddy(string p_initiator)
		{
			try
			{
				if(strBuddyID.Equals(p_initiator) || strBuddyID.Equals("NA"))
				{
					return true;
				}
				return false;
			}
			catch(Exception e)
 			{
                MessageBox.Show("Exception in isProperBuddy():"+e.Message);
				return false;
			}
		}
		private void UpdateSenderWindow(string p_RID)
		{
			try
			{
				if(SenderWindow.Contains(p_RID))
				{
					for(int i =0; i<SenderWindow.Count; i++)
					{
						if(SenderWindow[i].ToString().Equals(p_RID))
						{
							if(i==0)
							{
								SenderWindow.RemoveAt(0);
								while((SenderWindow.Count>0) &&
									SenderWindow[0].ToString().Equals("R"))
								{
									SenderWindow.RemoveAt(0);
								}
								break;
							}
							else 
							{
								SenderWindow.Insert(i,"R");
								SenderWindow.RemoveAt(i+1);
								break;
							}
						}						
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in UpdateSenderWindow(): "+e.Message);
			}
		}


	    private BroadcastPacket transformXMLIntoBroadcastPacket()
		{
			BroadcastPacket rcv_packet = new BroadcastPacket();
			try
			{				
				XmlDocument doc = new XmlDocument();
				doc.LoadXml(ReceivedData);
				ReceivedDataLength=0;//reset the received data's length
				XmlNode root = doc.DocumentElement; 
				XmlNodeList xNodes = root.ChildNodes;
			
				foreach(XmlNode childNode in xNodes)
				{					
					XmlElement elem = (XmlElement)childNode;
					if(elem.Name.Equals(Packet_CONSTANTS.TYPE))
					{
						rcv_packet.strType = elem.FirstChild.Value;
					}
					/*else if(elem.Name.Equals(Packet_CONSTANTS.CACHE))
					{
						rcv_packet.strOnCache = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.ACM))
					{
						rcv_packet.strOnACM = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.SWS))
					{
						rcv_packet.strSWS = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.RP))
					{
						rcv_packet.strOnRP = elem.FirstChild.Value;
					}*/	
					else if(elem.Name.Equals(Packet_CONSTANTS.CONF))
					{
                        rcv_packet.strOnConf = elem.FirstChild.Value;
					}
				}	
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception is occurred in transformXmlIntoBroadcastPacket() : "+e.Message);
			}
			return rcv_packet;					
		}

		private void SetAllBasicInfo(BroadcastPacket broadPacket)
		{
			try
			{
				//SWS_TB.Text = broadPacket.strSWS;
				/*if(broadPacket.strOnCache.Equals(true.ToString()))
				{
					Cache_CB.Checked = true;
				}
				else
				{
					Cache_CB.Checked = false;
				}
  
				if(broadPacket.strOnACM.Equals(true.ToString()))
				{
					ACM_CB.Checked = true;
				}
				else
				{
					ACM_CB.Checked = false;
				}
				if(broadPacket.strOnRP.Equals(true.ToString()))
				{
					RP_CB.Checked = true;
				}
				else
				{
					RP_CB.Checked = false;
				}*/
				if(broadPacket.strOnConf.Equals(true.ToString()))
				{
                    Conf_CB.Checked = true;
				}
				else
				{
					Conf_CB.Checked = false;
				}
				//OnReceiptPacket = RP_CB.Checked;
				//SWS = int.Parse(SWS_TB.Text);
				//OnCache = Cache_CB.Checked;
				//OnACM = ACM_CB.Checked;	
				UseOfConf = Conf_CB.Checked;		
				
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SetAllBasicInfo(): "+e.Message);
			}
		}


		private void DisplayMyChat(string p_message)
		{
            ChatBox.Text += strMyID+" says : "+p_message+"\r\n"; 
		}
		/// <summary>
		/// According to the Packet, each received packet is handled (receive/send) 
		/// appropriately.
		/// </summary>
		private void HandleReceivedPacket(object[] arguments)
		//private void HandleReceivedPacket()
		{
			try
			{				
				if(ReceivedDataLength>0)
				{
					if(isReceiptPacket())
					{						
						ReceiptPacket receivedPacket = transformXmlIntoReceiptPacket();						
						ReleaseReceiptTimer(receivedPacket.strReceiptID);
						UpdateReceiptBuffer(receivedPacket.strReceiptID);		
					}
					else if(isBroadcastNTPacket())
					{
						BroadcastPacket broadPacket = transformXMLIntoBroadcastPacket();
						startTickingNTTime();						
						SetAllBasicInfo(broadPacket);

						if(isSourceNode)//in source node
						{							
							Packet packet = new Packet();
							SetMySessionID();
							processedPacketSN = 0;
							SendStartChatPacket(ref packet); 
						}
					}
					else//All other packets
					{
						//SendReceiptPacket to previous node
						Packet receivedPacket = transformXmlIntoPacket();
						string previousNodeID = receivedPacket.strCurrentID;
						if(OnReceiptPacket)
						{
							SendReceiptPacket(receivedPacket, previousNodeID);
						}
							
						if(strMyID.Equals(receivedPacket.strTarget))//when current node is the target node
						{	
							//if the packet is not received yet, store it. if the packet is already received but
							//the path is new, then store it. Otherwise, do nothing.
							receivedPacket.strRouteRecord = receivedPacket.strRouteRecord+":"+strMyID;
							StoreInCacheOfTargetNode(receivedPacket);//store in cache if necessary

							//debugging
							/*if(!receivedPacket.strType.Equals(Packet_CONSTANTS.START_CHAT_PACKET) &&
								!receivedPacket.strType.Equals(Packet_CONSTANTS.REJECT_START_CHAT_PACKET) &&
								!receivedPacket.strType.Equals(Packet_CONSTANTS.ACCEPT_START_CHAT_PACKET))
							{
								ChatBox.Text = ChatBox.Text+"\r\n RCVD: "+ receivedPacket.strInitiator+": "+
									receivedPacket.strTimeStamp+" "+
									receivedPacket.strType+" "+receivedPacket.strRequestID+
									" "+receivedPacket.strRouteCache; 
							}*/

							if(!isAlreadyReceived(receivedPacket))
							{
								//1. Received "Start Chat Message"
								if(receivedPacket.strType.Equals(Packet_CONSTANTS.START_CHAT_PACKET))
								{
			//						MessageBox.Show("SCP received");

									if(isProperBuddy(receivedPacket.strInitiator))
									{
										isDoneConnectionEstablishment = true;       
										SetBuddySessionID(receivedPacket);               
										strBuddyID = receivedPacket.strInitiator;
										strBuddyIP = GetIPAddressByIDInConfFile(strBuddyID);	
										answer = MessageBox.Show(receivedPacket.strInitiator+" wants to chat with you! ",
											receivedPacket.strType,
											MessageBoxButtons.YesNo,
											MessageBoxIcon.Question,
											MessageBoxDefaultButton.Button1).ToString();
						                
										if(answer.Equals("Yes"))
										{
											acceptedPacket = true;
											showChatPanel();			
											SetMySessionID();
										}								
										SendReplyStartChatPacket(ref receivedPacket);				
										processedPacketSN = getSequenceNumFromRequestIDReceived(receivedPacket.strRequestID);									
										if(answer.Equals("Yes"))
										{
											BuddyID_IP.Text = strBuddyID+" :"+strBuddyIP;
											MyID_IP.Text = strMyID+" :"+strMyIP;
										}
									}
									else//not proper buddy
									{
										OnCallingBuddy.Add(receivedPacket.strRequestID);
										SendOnCallingPacket(ref receivedPacket);                       
									}
								}
									//2. Received "Hang Up Message"  
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.HANG_UP_PACKET))
								{
									isDoneConnectionEstablishment = false;
									receivedPacket.strRouteCache = getReverseRoute(receivedPacket.strRouteCache);									
									SendRouteReplyPacket(ref receivedPacket);
    								ChatBox.Text = strBuddyID+" hung up the conversation !";
									Thread.Sleep(2000);		
									showBuddyScreen();
									resetAll();									
								}
									//3. "Data Message"							
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.DATA_PACKET))
								{
									AddToReceiverWindow(receivedPacket);
									SetBuddySessionID(receivedPacket);
									receivedPacket.strRouteCache = getReverseRoute(receivedPacket.strRouteCache);
									SendRouteReplyPacket(ref receivedPacket);
									ProcessPacketsInReceiverWindow();
								}
									//4.  Received "Reject Start Chat Message"  as Route Reply Packet
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.REJECT_START_CHAT_PACKET))
								{
									ReleaseRRTimer(receivedPacket.strRequestID);
									UpdateSendBuffer(receivedPacket.strRequestID);
									string initiator = receivedPacket.strInitiator;	
									MessageBox.Show(initiator+" disagrees to talk with you", "Accepting Chatting With You!",
										MessageBoxButtons.OK,MessageBoxIcon.Exclamation,MessageBoxDefaultButton.Button1);
									resetAll();
								}
									//5.  Received "Accept Start Chat Message"  as Route Reply Packet
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.ACCEPT_START_CHAT_PACKET))
								{			
									BuddyID_IP.Text = strBuddyID+":"+strBuddyIP;
									MyID_IP.Text = strMyID+":"+strMyIP;
									ReleaseRRTimer(receivedPacket.strRequestID);
									UpdateSendBuffer(receivedPacket.strRequestID);
									string initiator = receivedPacket.strInitiator;										
									MessageBox.Show(initiator+" agrees to talk with you", "Accepting Chatting With You!",
										MessageBoxButtons.OK,MessageBoxIcon.Exclamation,MessageBoxDefaultButton.Button1);
                                    isDoneConnectionEstablishment = true;
									showChatPanel();
								}
									//6. Received "Route Reply Message" 
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.ROUTE_REPLY_PACKET))									
								{
									UpdateSenderWindow(receivedPacket.strRequestID);
									ReleaseRRTimer(receivedPacket.strRequestID);
									if(send_buffer.Contains(receivedPacket.strRequestID))
									{
										Packet temp = getLostPacketFromSendBuffer(receivedPacket.strRequestID);
										string message = temp.strPayload;
										DisplayMyChat(message);                                        
									}
									UpdateSendBuffer(receivedPacket.strRequestID);
								}							
								   //7. "Reply Hang Up Message"
								if(receivedPacket.strType.Equals(Packet_CONSTANTS.REPLY_HANG_UP_PACKET))
								{
									ReleaseRRTimer(receivedPacket.strRequestID);
									UpdateSendBuffer(receivedPacket.strRequestID);								
									resetAll();
								}
									//8. Received "Route Error Message" 
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.ROUTE_ERROR_PACKET))
								{
									ReleaseRRTimer(receivedPacket.strRequestID);
									
									if(send_buffer.Contains(receivedPacket.strRequestID))
									{
										Packet packet = getLostPacketFromSendBuffer(receivedPacket.strRequestID);
										if(routeCache.hasPath(packet.strRouteCache) )
										{
											routeCache.RemovePath(packet.strRouteCache);
										}				
										RepairRouteError(ref receivedPacket);
									}
								}
									//9. Received "Route Discovery Packet"
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.ROUTE_DISCOVERY_PACKET))
     							{
									SetBuddySessionID(receivedPacket);
									receivedPacket.strRouteCache = getReverseRoute(receivedPacket.strRouteRecord);
									SendRouteReplyPacket(ref receivedPacket);
								}
									//10. Received "Reply Route Discovery Packet"
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.REPLY_ROUTE_DISCOVERY_PACKET))
								{
									ReleaseRRTimer(receivedPacket.strRequestID);
									UpdateSendBuffer(receivedPacket.strRequestID);
									processingRDBuffer.Remove(receivedPacket.strPayload);
									string lostPacketRequestID =receivedPacket.strPayload;
									if(send_buffer.Contains(lostPacketRequestID) && 
										!RRTimer2Buffer.Contains(lostPacketRequestID))
									{										
										Packet packetInSB = getLostPacketFromSendBuffer(lostPacketRequestID);
										string timerID = packetInSB.strRequestID;
										Packet packetToResend = createPacketForRetransmission(ref packetInSB);
										packetToResend.strRouteCache = getReverseRoute(receivedPacket.strRouteCache); 
										UpdateSendBuffer(packetInSB.strRequestID);//remove old DP
										send_buffer.Add(packetToResend.strRequestID, packetToResend);//add new DP with new routeCache
										SetRouteReplyTimer(timerID, packetToResend);									
										SendPacket(ref packetToResend);                                    
									}
								}
									//11. Received "OnCalling Packet"
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.ONCALLING_PACKET))
								{
									ReleaseRRTimer(receivedPacket.strRequestID);
									UpdateSendBuffer(receivedPacket.strRequestID);
									resetAll();
									//Thread.Sleep(1000);
									MessageBox.Show(receivedPacket.strInitiator+" is busy!.");										
								}
								//}
							}// not aready received						
							else//the packet is already received.: For debugging purpose
							{								
								if(receivedPacket.strType.Equals(Packet_CONSTANTS.START_CHAT_PACKET))
								{
									SendReplyStartChatPacket(ref receivedPacket);
								}
								else if(receivedPacket.strType.Equals(Packet_CONSTANTS.DATA_PACKET) ||
									receivedPacket.strType.Equals(Packet_CONSTANTS.HANG_UP_PACKET))
								{
									receivedPacket.strRouteCache = getReverseRoute(receivedPacket.strRouteRecord);
									SendRouteReplyPacket(ref receivedPacket);
								}								
							}//already received packet			
						}//if current node is target
						else//when current node is not the target node
						{
							if(receivedPacket.strType.Equals(Packet_CONSTANTS.HANG_UP_PACKET))
							{
                                resetAll();
							}
							ForwardPacketToNextNode(ref receivedPacket);
						}

					}//not receipt packet									
				}//received data > 0
			}//try
			catch(Exception e)
			{
				MessageBox.Show("Exception is HandleReceivedPacket(): "+e.Message);
				MessageBox.Show("ReceivedData: "+ReceivedData);
			}
			//MessageBox.Show("end of handle received packet()");
		}
		
		private void AddToReceiverWindow(Packet p_packet)
		{
			try
			{
				if(ReceiverWindow.Count<SWS)
				{
					int SN = getSequenceNumFromRequestIDReceived(p_packet.strRequestID);
					ElementsForReceiverWindow ERW = new ElementsForReceiverWindow();
					ERW.SN = SN.ToString();
					ERW.Payload = p_packet.strPayload;
					InsertInProperPositionInRW(ERW);
				}
					
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in AddToReceiverWindow(): "+e.Message);
			}
		}

		private void InsertInProperPositionInRW(ElementsForReceiverWindow p_ERW)
		{
			try
			{
				int receivedSN = int.Parse(p_ERW.SN);
				bool isInserted = false;
				for(int i =0; i<ReceiverWindow.Count; i++)
				{
					ElementsForReceiverWindow ERW = (ElementsForReceiverWindow)ReceiverWindow[i];
					int tempSN = int.Parse(ERW.SN);
					if(receivedSN<tempSN)
					{					
						ReceiverWindow.Insert(i,p_ERW);
						isInserted = true;
						break;
					}
				}
				if(!isInserted)
				{
					ReceiverWindow.Add(p_ERW);
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in InsertInProperPositionInRW(): "+e.Message);
			}
		}

		private Packet createPacketForRetransmission(ref Packet p_packet)
		{
			Packet packet = new Packet();
			try
			{
                packet.strInitiator = p_packet.strInitiator;
			    packet.strTarget = p_packet.strTarget;
                packet.strRequestID = p_packet.strRequestID;
				packet.strCurrentID = strMyID;
				packet.strPayload = p_packet.strPayload;
				packet.strRouteRecord = strMyID;
				packet.strType = p_packet.strType;
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in createPacketForRetransmission(): "+e.Message);
			}
			return packet;
		}


		private void SetBuddySessionID(Packet p_packet)
		{
			if(BuddySessionID.Equals(""))
			{
				strBuddyID = p_packet.strInitiator;
				BuddySessionID = getSessionIDFromRequestID(p_packet.strRequestID); 
			}
		}
		/// <summary>
		/// Show the data in order from ReceiverBuffer
		/// </summary>
		/// <summary>
		/// Show the data in order from ReceiverWindow
		/// </summary>
		public void ProcessPacketsInReceiverWindow()
		{
			try
			{								
				while(ReceiverWindow.Count>0) 
				{
					int nextPacketSN = processedPacketSN+1;
					if(isInReceiverWindow(nextPacketSN.ToString()))
					{
						ElementsForReceiverWindow ERW = 
							(ElementsForReceiverWindow)ReceiverWindow[0];
						if(ERW.SN.Equals(nextPacketSN.ToString()))
						{
							processedPacketSN = nextPacketSN;
							string message = getPayloadFromReceiverWindow(nextPacketSN.ToString());
							//In prototype, you need to display the message
							DisplayBuddyChat(message);
							ReceiverWindow.RemoveAt(0);
						}
						else
						{
							MessageBox.Show("Hello something wrong in ProcessPacketsInReceiverWindow()");
						}
					}
					else
					{
						break;
						//Need to wait until the nextPacketSN is available 
						//by seeing the next arriving packet
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in ProcessPacketsInReceiverWindow() : "+e.Message);
			}
		}


		public int getSequenceNumFromRequestIDReceived(string p_requestID)
		{
			string requestID = p_requestID;
			string SID="";
			try
			{
				SID = getSessionIDFromRequestID(p_requestID);
				string strSeq = p_requestID.Substring(SID.Length+1);
				int new_seq = int.Parse(strSeq);
				return new_seq;
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception in getSequenceNumFromRequestIDReceived(): "+e.Message+" requestID is "+requestID+" SID is "+SID);
				return -100;
			}
		}



		/// <summary>
		/// Receiver node needs to get the SessionID which is determined by the sender node.
		/// </summary>
		/// <param name="p_requestID"></param>
		/// <returns></returns>
		private string getSessionIDFromRequestID(string p_requestID)
		{
			string SID = "";
			try
			{
				char colon = ':';
				int numOfColon = 0;
				for(int i = 0; i<p_requestID.Length; i++)
				{
					if(p_requestID[i].Equals(colon))
					{
						numOfColon++;
					}
					if(numOfColon <3)
					{
						SID = SID+p_requestID[i].ToString();
					}
					else
					{
						break;
					}

				}
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in getSessionIDFromRequestID(): "+e.Message);
			}

			return SID;
		}

		private void DisplayBuddyChat(string message)
		{
			ChatBox.Text += strBuddyID+" says : "+message+"\r\n";
		}

		public string getPayloadFromReceiverWindow(string p_SN)
		{
			string payload = "N/A";
			try
			{
				IEnumerator ie = ReceiverWindow.GetEnumerator();
				while(ie.MoveNext())
				{
					ElementsForReceiverWindow ERW = (ElementsForReceiverWindow)ie.Current;
					if(ERW.SN.Equals(p_SN))
					{
						payload = ERW.Payload;
						break;
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getPacketFromReceiverWindow(): "+e.Message);
			}
			return payload;
		}

		private string showSendBufferInside()
		{
			string sendBufferInside = "";
			IDictionaryEnumerator ide = send_buffer.GetEnumerator();
			while(ide.MoveNext())
			{
				Packet temp_packet = (Packet)ide.Value;
				sendBufferInside = sendBufferInside+"\r\n"+ide.Key.ToString()+" "+temp_packet.strType;
			}
			if(sendBufferInside.Equals(""))
			{
				sendBufferInside = "NA";
			}
			return sendBufferInside;			 
		}

		
		

		/// <summary>
		/// Send a receipt packet to the previous node unless the packet's type is
		/// not a receipt packet.
		/// </summary>
		/// <param name="packet">packet</param>
		/// <param name="p_previousNodeID">previous node</param>
		private void SendReceiptPacket(Packet packet, string p_previousNodeID)
		{
			try
			{
				if(packet.strType.Equals(Packet_CONSTANTS.DATA_PACKET) ||
					packet.strType.Equals(Packet_CONSTANTS.ROUTE_REPLY_PACKET))
				{

					ReceiptPacket rctPacket = new ReceiptPacket();
					rctPacket.strType = Packet_CONSTANTS.RECEIPT_PACKET;
					rctPacket.strReceiptID = packet.strReceiptID;
                
					XmlDocument doc = new XmlDocument();
					doc.LoadXml("<Packet></Packet>");

					XmlNode root = doc.DocumentElement;

					//1. Type
					XmlElement elem = doc.CreateElement("Type");
					elem.InnerText = rctPacket.strType;
					root.AppendChild(elem);

					//2.
					elem = doc.CreateElement("ReceiptID");
					elem.InnerText = rctPacket.strReceiptID;
					root.AppendChild(elem);

					string messageToSend = doc.InnerXml.ToString();
					string serverID = p_previousNodeID;
					neighbors = getNeighbors(strMyID);					
					if(neighbors.Contains(serverID))
					{
						string serverIP = GetIPAddressByIDInConfFile(serverID);                        
						bool sentSuccess = sendUDP(serverIP, messageToSend);
						//			MessageBox.Show("sentSuccess for receipt packet : "+rctPacket.strReceiptID+sentSuccess.ToString());
					}
					else
					{
						//MessageBox.Show(strMyID+"-"+serverID+" link for Route Error"); 
					}
				}
			}
			catch(Exception ex)
			{
				MessageBox.Show("Exception is occurred in SendReceiptPacket(): "+ex.Message);
			}
		}

		
		private void SendRouteReplyPacket(ref Packet packet)
		{
			try
			{
				if(packet.strType.Equals(Packet_CONSTANTS.ROUTE_DISCOVERY_PACKET))
				{
					packet.strType = Packet_CONSTANTS.REPLY_ROUTE_DISCOVERY_PACKET;				
				}
				else if(packet.strType.Equals(Packet_CONSTANTS.HANG_UP_PACKET))
				{
					packet.strType = Packet_CONSTANTS.REPLY_HANG_UP_PACKET;
				}
				else
				{
					packet.strType = Packet_CONSTANTS.ROUTE_REPLY_PACKET;
				}

				//MessageBox.Show("Send route reply packet");
				string target = packet.strInitiator;
			
				if(packet.strRouteCache.Equals(Packet_CONSTANTS.NOTAVAILABLE))
				{
					if(!routeCache.isEmpty())
					{
						packet.strRouteCache = routeCache.getPath();
					}
				}
				packet.strInitiator = strMyID;
				packet.strTarget = target;
				packet.strCurrentID = strMyID;
				if(!packet.strType.Equals(Packet_CONSTANTS.REPLY_ROUTE_DISCOVERY_PACKET))
				{
					packet.strPayload = Packet_CONSTANTS.NOTAVAILABLE;
				}
				//packet's request ID remains same. it will be checked in the original initiator
				packet.strRouteRecord = strMyID;
				SendPacket(ref packet);
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception occurred in SendRouteReplyPacket(): "+e.Message);
			}
		}

		private string getReceiptID(string p_currentID, string p_nextID)
		{            
			string receiptID = "";
			Random receiptNum = new Random();
			int id = -1;
			while(true)
			{
				id = receiptNum.Next(100000);
				receiptID = p_currentID+":"+p_nextID+":"+id.ToString();
				if(!ReceiptBuffer.ContainsKey(receiptID))
				{
					break;  
				}
			}
			return receiptID;
		}

		private void SendStartChatPacket(ref Packet packet)
		{
			try
			{
				packet.strCurrentID = strMyID;
				packet.strInitiator = strMyID;
				packet.strPayload = Packet_CONSTANTS.NOTAVAILABLE;
				packet.strTarget = strBuddyID;
				packet.strType = Packet_CONSTANTS.START_CHAT_PACKET;
				packet.strRouteRecord = strMyID;
				if(!routeCache.isEmpty())
				{
					packet.strRouteCache = routeCache.getPath();
				}
				else
				{
					packet.strRouteCache = Packet_CONSTANTS.NOTAVAILABLE;
				}		    
				packet.strRequestID = getRequestID(packet.strType);
				SaveInSendBuffer(packet);
				string timerID = packet.strRequestID;
    			SetRouteReplyTimer(timerID, packet);
				SendPacket(ref packet);	
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SendStartChatPacket() : "+e.Message);
			}
		}

		private void SendMessagePacket(ref Packet packet)
		{
			try
			{
				packet.strInitiator = strMyID;
				packet.strCurrentID = strMyID;
				packet.strPayload = strMessageBox;
				packet.strTarget = strBuddyID;
				if(!routeCache.isEmpty())
				{
					packet.strRouteCache = routeCache.getPath();	
				}
				else
				{
					packet.strRouteCache = Packet_CONSTANTS.NOTAVAILABLE;			
				}
				packet.strRouteRecord = strMyID;
				packet.strType = Packet_CONSTANTS.DATA_PACKET;
				packet.strRequestID = getRequestID(packet.strType);
				if(SenderWindow.Count<SWS)
				{
					SenderWindow.Add(packet.strRequestID);
					SaveInSendBuffer(packet);
					string timerID = packet.strRequestID;
					if(!RRTimer2Buffer.Contains(timerID))
					{
						SetRouteReplyTimer(timerID, packet);
						SendPacket(ref packet);	
					}
				}
				else
				{
				//	MessageBox.Show("SendMessagePacket(): Please slow down to send messages.");
					ChatBox.Text += "\r\n Please slow down to send messages.";
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SendMessagePacket():"+e.Message);
			}
		}

		private Packet getPacketInSendBuffer()
		{
			Packet packet = new Packet();
			IEnumerator ie = send_buffer.Keys.GetEnumerator();
			string smallestRID = "";
			while(ie.MoveNext())
			{
				string comp = ie.Current.ToString();
				if(smallestRID.Equals(""))
				{
					smallestRID = comp;
				}
				else
				{
					if(smallestRID.CompareTo(comp)>0)
					{
						smallestRID = comp;
					}
				}				           
			}
			IDictionaryEnumerator ide = send_buffer.GetEnumerator();
			while(ide.MoveNext())
			{
				if(ide.Key.ToString().Equals(smallestRID))
				{
					packet = (Packet)ide.Value;
					break;
				}
			}
			return packet;
		}

		private void SendHangUpPacket(ref Packet packet)
		{
			try
			{
				packet.strType = Packet_CONSTANTS.HANG_UP_PACKET;
				packet.strCurrentID = strMyID;
				packet.strInitiator = strMyID;
				packet.strPayload = Packet_CONSTANTS.NOTAVAILABLE;
				packet.strTarget = strBuddyID;
				packet.strRouteRecord = strMyID;
				if(!routeCache.isEmpty())
				{
					packet.strRouteCache = routeCache.getPath();
				}
				else
				{
					packet.strRouteCache = Packet_CONSTANTS.NOTAVAILABLE;
				}		
				packet.strRequestID = getRequestID(packet.strType);				
				
				while(true)
				{
					if(SenderWindow.Count==0)
					{
						SaveInSendBuffer(packet);
						string timerID = packet.strRequestID;
						SetRouteReplyTimer(timerID, packet);
						SendPacket(ref packet);	
						break;
					}
					else
					{
						Thread.Sleep(1000);
					}			
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SendHangUpPacket():"+e.Message);
			}
		}

/*		private void SetRouteReplyTimer(string p_ID, Packet p_packet)
		{
			//for connection establishment packet, allow 5 seconds to get 
			//the answer for the chat request

			int time_interval =0;
			if(p_packet.strType.Equals(Packet_CONSTANTS.START_CHAT_PACKET)||
				p_packet.strType.Equals(Packet_CONSTANTS.ROUTE_DISCOVERY_PACKET))
			{
				time_interval = 5000;
			}
			else
			{
				time_interval = 2000;
			}
			MyTimer myTimer = new MyTimer(p_ID, time_interval,this,p_packet);			              
			myTimer.setRRTimeoutTimer();
			RRTimerBuffer.Add(p_ID, myTimer);			
			
		}
		*/

		/// <summary>
		/// With new Timer2
		/// </summary>
		/// <param name="p_ID"></param>
		/// <param name="p_packet"></param>
		private void SetRouteReplyTimer(string p_ID, Packet p_packet)
		{
			try
			{
				int time_interval =0;
				time_interval = 4000;
				MyTimer2 myTimer = new MyTimer2(time_interval,true,3,this,p_packet,p_ID);
				myTimer.SetRRTimer2();
				RRTimer2Buffer.Add(p_ID, myTimer);
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception in SetRouteReplyTimer(): "+e.Message+" id is "+p_ID+" packet's type is "+p_packet.strType);
			}            
		}


		private void SendReplyStartChatPacket(ref Packet packet)
		{
			try
			{
				string initiator = packet.strTarget;
				string target = packet.strInitiator;

				packet.strInitiator = initiator;
				packet.strTarget = target;
				packet.strRouteCache = getReverseRoute(packet.strRouteRecord);
				packet.strRouteRecord = strMyID;
				packet.strCurrentID = strMyID;
				packet.strPayload = Packet_CONSTANTS.NOTAVAILABLE;
   
				if(acceptedPacket)
				{
					packet.strType = Packet_CONSTANTS.ACCEPT_START_CHAT_PACKET;  
				}
				else
				{
					packet.strType = Packet_CONSTANTS.REJECT_START_CHAT_PACKET;  
				}			
				SendPacket(ref packet);						
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in SendReplyStartChatPacket(): "+e.Message);
			}
		}
		
	
		/// <summary>Debug is done
		/// If the received packet is a reply packet by checking an original request ID 
		/// in send buffer, the received reply packet is removed in the send buffer.
		/// </summary>
		/// <param name="p_RequestID"></param>
		public void UpdateSendBuffer(string p_RequestID)
		{
			try
			{
				if(send_buffer.Contains(p_RequestID))
				{
					send_buffer.Remove(p_RequestID);
				}				
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception ocurred in UodateSendBuffer() : "+e.Message);
			}
		}

		public void ReleaseRRTimer(string p_RequestID)
		{
			/*
			IDictionaryEnumerator ide = RRTimerBuffer.GetEnumerator();
			while(ide.MoveNext())
			{
				if(ide.Key.ToString().Equals(p_RequestID))
				{
					MyTimer tempTimer = (MyTimer)ide.Value;
					tempTimer.stopMyTimer();
					RRTimerBuffer.Remove(p_RequestID);
					//MessageBox.Show("Released RRTimer");
					break;
				}
			}*/
			try
			{
				IDictionaryEnumerator ide = RRTimer2Buffer.GetEnumerator();
				while(ide.MoveNext())
				{
					if(ide.Key.ToString().Equals(p_RequestID))
					{
						MyTimer2 tempTimer = (MyTimer2)ide.Value;
						tempTimer.setEnabled(false);
						break;
					}
				}
				RRTimer2Buffer.Remove(p_RequestID);
			}
			catch(Exception e)
			{
                 MessageBox.Show("Exception in ReleaseRRTimer(): "+e.Message);
			}
		}


		/// <summary>
		/// When the current node is a target node, route cache is set appropriately
		/// : route destination and path in cache of current node
		/// </summary>
		/// <param name="packet">received packet</param>
		private void StoreInCacheOfTargetNode(Packet packet)
		{
			try
			{
				if(packet.strTarget.Equals(strMyID))
				{
					if(isProperBuddy(packet.strInitiator))//(isProperBuddy(packet.strRequestID))
					{
						if(!(packet.strType.Equals(Packet_CONSTANTS.ROUTE_ERROR_PACKET)) &&
							!(packet.strType.Equals(Packet_CONSTANTS.ONCALLING_PACKET)))
						{
							if(packet.strType.Equals(Packet_CONSTANTS.ROUTE_DISCOVERY_PACKET) ||
								packet.strType.Equals(Packet_CONSTANTS.REPLY_ROUTE_DISCOVERY_PACKET))
							{
								if(!isAlreadyReceived(packet))
								{
									ChatBox.Text += "route Cache cleared!";
									routeCache.Clear();
								}
							}
							//SetCache();
							string reverseRoute  = getReverseRoute(packet.strRouteRecord);
							routeCache.AddPath(reverseRoute, OnCache);
						}
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception occurred in StoreInCacheOfTargetNode() : "+e.Message);
			}
		}

		

		/// <summary>
		/// Return a reverse route for entered route. Debug is done!
		/// </summary>
		/// <param name="route">Route</param>
		/// <returns></returns>
		private string getReverseRoute(string route)
		{
			char[] temp = route.ToCharArray();
			char[] reverseRoute = new char[temp.Length];
			int length = temp.Length-1;
			for(int i=0; i<reverseRoute.Length; i++)
			{
				reverseRoute[i] = temp[length-i];
				//Console.WriteLine("reverseRoute : "+reverseRoute[i]);
			}
			string strRvsRoute = "";
			for(int i=0; i<reverseRoute.Length; i++)
			{
				strRvsRoute = strRvsRoute+reverseRoute[i].ToString();	
			}                
			return strRvsRoute;
		}

		/// <summary>
		/// In forwarding a packet, get the previous node ID.
		/// </summary>
		/// <param name="p_RouteRecord"></param>
		/// <returns>string which is a previous node ID</returns>
		private string getPreviousNode(string p_RouteRecord)
		{
			for(int i=2; i<p_RouteRecord.Length; i++)
			{
				if(strMyID.Equals(p_RouteRecord[i].ToString()))
				{
					return p_RouteRecord[i-2].ToString(); 
				}
			}
			return null;
		}

		private void ForwardPacketToNextNode(ref Packet packet)
		{
			packet.strRouteRecord = packet.strRouteRecord+":"+strMyID;
			packet.strCurrentID = strMyID;

//			string reversePath = getReverseRoute(packet.strRouteRecord);
//			routeCache.AddPath(packet.strInitiator, reversePath);
			
			/*if(packet.strType.Equals(Packet_CONSTANTS.ROUTE_ERROR_PACKET))
			{
				   UpdateRouteCache(packet);
			}
			else
			{*/
/*			if(packet.strRouteCache.Equals(Packet_CONSTANTS.NOTAVAILABLE))
			{   
				if(routeCache.hasDestination(packet.strTarget))
				{
					packet.strRouteCache = routeCache.getPath(packet.strTarget);      
				}
			}*/
			//}
			SendPacket(ref packet);
		}


		/// <summary>
		/// After receiving the packet, transform the Xml file to Packet object.
		/// </summary>
		/// <returns>Packet Object containing all information of xml data</returns>
		private Packet transformXmlIntoPacket()
		{
			/*if(ReceivedData.StartsWith("<Packet><Type>"+Packet_CONSTANTS.ROUTE_DISCOVERY_PACKET))
			{
				MessageBox.Show("transformXmlIntoPacket(): "+ReceivedData);
			}*/
			Packet rcv_packet = new Packet();
			try
			{
				////Monitor.Enter(this);
				XmlDocument doc = new XmlDocument();
				//P();//Semaphore P
				doc.LoadXml(ReceivedData);
				ReceivedDataLength=0;//reset the received data's length
				//V();//Semaphore V
				XmlNode root = doc.DocumentElement; 
				XmlNodeList xNodes = root.ChildNodes;
			
				foreach(XmlNode childNode in xNodes)
				{
					
					XmlElement elem = (XmlElement)childNode;
					if(elem.Name.Equals(Packet_CONSTANTS.TYPE))
					{
						rcv_packet.strType = elem.FirstChild.Value;						
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.TARGET))
					{
						rcv_packet.strTarget = elem.FirstChild.Value;
					}
					else if (elem.Name.Equals(Packet_CONSTANTS.PAYLOAD))
					{
						rcv_packet.strPayload = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.CURRENTID))
					{
						rcv_packet.strCurrentID = elem.FirstChild.Value; 

					}
					else if(elem.Name.Equals(Packet_CONSTANTS.INITIATOR))
					{
						rcv_packet.strInitiator = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.ROUTECACHE))
					{
						rcv_packet.strRouteCache = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.REQUESTID))
					{
						rcv_packet.strRequestID = elem.FirstChild.Value;

					}
					else if(elem.Name.Equals(Packet_CONSTANTS.ROUTERECORD))
					{
						rcv_packet.strRouteRecord = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.RECEIPTID))
					{
						rcv_packet.strReceiptID = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(Packet_CONSTANTS.TIMESTAMP))
					{
						rcv_packet.strTimeStamp = elem.FirstChild.Value;
					}
				}	
				////Monitor.Exit(this);
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception is occurred in transformXmlIntoPacket() : "+e.Message);
			}
			return rcv_packet;			
		}

		/// <summary>
		/// Get the next node in the strRouteCache
		/// </summary>
		/// <param name="p_path">appropriate path is passed</param>
		/// <returns>the next node to go</returns>
		private string getNextNodeInCachePath(string path)
		{

			string currentNode = strMyID;
			try
			{
				int index = path.IndexOf(currentNode);
				//MessageBox.Show("index = "+index);
				int nextNodeStartIndex = index+currentNode.Length+1;
				//MessageBox.Show("nextNodeStartIndex = "+nextNodeStartIndex);
				string subStr = path.Substring(nextNodeStartIndex);
				//MessageBox.Show("subStr: "+subStr);
			
				int colon = subStr.IndexOf(":");
				if(colon>0)
				{
					return subStr.Substring(0, colon);
				}
				else
				{
					return subStr;
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getNextNodeInCachePath():"+e.Message+" path: "+path); 
				return null;
			}
			  
			
		}

		/// <summary>
		/// The nodes read from conf.xml are shown in the buddy list(listView)
		/// </summary>
		private void showBuddyList()
		{
			try
			{
				BuddyList.View = View.Details;
				BuddyList.FullRowSelect = true;
				BuddyList.Columns.Add("Buddy ID",-2,HorizontalAlignment.Left);
				BuddyList.Columns.Add("IP Address", BuddyList.Width, HorizontalAlignment.Left);
				IDictionaryEnumerator ide = hashBuddyList.GetEnumerator();						
				while(ide.MoveNext())
				{
					if(!ide.Key.ToString().Equals(strMyID))
					{

						ListViewItem id = new ListViewItem(ide.Key.ToString());
						id.SubItems.Add(ide.Value.ToString());
						BuddyList.Items.Add(id);
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("showBuddyList() Exception occurred: "+e.Message);  
			}
		}

		/// <summary>
		/// In Hashtable hashSequence, source as Key and sequence as Value are stored.
		/// </summary>
		public void storeSequenceInHashtable()
		{
			try
			{
				XmlNode root = xmlDoc.DocumentElement;//configuration root element
				XmlNodeList xNodes = root.ChildNodes;
											
				foreach(XmlNode childNode in xNodes)//<Node>
				{
					XmlElement currentElement = (XmlElement)childNode;
					//add to the hashSequence
					string source = currentElement.GetAttribute("ID").ToString();
					XmlNodeList SequenceNodes = childNode.ChildNodes;
					ArrayList temp_sequence = new ArrayList();
					foreach(XmlNode sequenceNode in SequenceNodes)//<Sequence>
					{
						XmlNodeList hopNodes = sequenceNode.ChildNodes; 

						foreach(XmlNode hopNode in hopNodes)//<Hop>
						{
							XmlNodeList pathNodes = hopNode.ChildNodes;
							Sequence s = new Sequence();
							foreach(XmlNode pathNode in pathNodes)
							{

								XmlElement cn = (XmlElement)pathNode;
								if(cn.Name.Equals("Destination"))
								{
									string temp = cn.FirstChild.Value;

									Point p = new Point();
									p.row = getRow(temp);
									p.col = getCol(temp);
									s.destination = p;
								}
								else if(cn.Name.Equals("Speed"))
								{
									s.speed = Convert.ToInt32(cn.FirstChild.Value);
								}
								else//Pause
								{
									s.pause = Convert.ToInt32(cn.FirstChild.Value);
								}
							}
							temp_sequence.Add(s);
						}
					}//outer foreach
					hashSequence.Add(source,temp_sequence);
				}									               
			}
			catch(Exception ex)
			{
				MessageBox.Show("Exception in storeSequenceInHashtable():"+ex.Message);
			}   
		}
 

		private void showHashSequence()
		{
			try
			{
				string inside ="HashSequence: \r\n";
				IDictionaryEnumerator ide = hashSequence.GetEnumerator();   
				while(ide.MoveNext())
				{
					string strSeq = "";
					inside = inside +ide.Key.ToString()+" :\r\n";
					ArrayList temp_al = (ArrayList)ide.Value;
					for(int i = 0; i<temp_al.Count; i++)
					{
						Sequence s = (Sequence)temp_al[i];						
						strSeq =strSeq+"D = ("+s.destination.row+","+s.destination.col+") S= "+
							s.speed+" P= "+s.pause+"\r\n";							
					}
					inside = inside+strSeq;
				}
				MessageBox.Show(inside);
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in showHashSequence() :"+e.Message);
			}
		}

		/// <summary>
		/// When string (row,col) is passed as a parameter, return row as an integer
		/// </summary>
		/// <param name="s">destination</param>
		/// <returns>row as an integer</returns>
		public int getRow(string s)
		{
			try
			{
				int leftP = s.IndexOf("(");		
				int dot = s.IndexOf(",");
				string strRow = s.Substring(leftP+1, dot-1);
				//MessageBox.Show("strRow : "+strRow);
				int row = Convert.ToInt32(strRow);
				return row;
			}
			catch(Exception e)
			{
				MessageBox.Show("getRow() Exception occurred :"+e.Message);                
				return -1;
			}
		}

		/// <summary>
		/// When string (row,col) is passed as a parameter, return col as an integer
		/// </summary>
		/// <param name="s">destination</param>
		/// <returns>col as an integer</returns>
		public int getCol(string s)
		{
			try
			{
				int rightP = s.IndexOf(")");
				int dot = s.IndexOf(",");
				int left = dot+1;
				int right =rightP-left;
				//MessageBox.Show("left : "+left+" right :"+right);
				string strCol = s.Substring(dot+1, right);
				//MessageBox.Show("strCol : "+strCol);
				int col = Convert.ToInt32(strCol);
				return col;
			}
			catch(Exception e)
			{
				MessageBox.Show("getCol() Exception occurred :"+e.Message);                
				return -1;
			}
		}

		/// <summary>
		/// Return an integer for the randomly selected sequence number 
		/// between 0 and 100000
		/// </summary>
		/// <returns></returns>
		private int getSessionID()
		{
			Random random = new Random();
			int random_seuqnce_num = random.Next(0, 100000);
			return random_seuqnce_num;
		}

		/// <summary>
		/// Return a request ID as a format: "session id : initiator: target: sequenceNum" 
		/// </summary>
		/// <param name="p_target"></param>
		/// <returns></returns>
		private string getRequestID(string p_type)
		{
			string requestID = "";
			if(p_type.Equals(Packet_CONSTANTS.START_CHAT_PACKET))
			{
				requestID = MySessionID+":"+"0";				
			}
			else if(p_type.Equals(Packet_CONSTANTS.HANG_UP_PACKET))
			{
				requestID = MySessionID+":"+int.MaxValue.ToString();				
			}
			else if(p_type.Equals(Packet_CONSTANTS.ROUTE_DISCOVERY_PACKET))
			{
				discoverSeqNum--;
				requestID = MySessionID+":"+discoverSeqNum;				
			}
			else if(p_type.Equals(Packet_CONSTANTS.DATA_PACKET))
			{
				sequenceNum++;
				requestID = MySessionID+":"+sequenceNum;  
			}
			return requestID;		
		}
	
		/// <summary>
		/// Main launch point of the application
		/// </summary>
		/// <param name="args">argument string</param>
		static void Main(string[] args)
		{		
			Application.Run(new PocketPCChat());
			//Application.Exit();
		}


		//UdpServer related methods
		public void start_server()
		{	
			try
			{		
				//Starting the UDP Listening thread.
				udpListeningThread = new Thread(new ThreadStart(this.receivePacket));
				udpListeningThread.Start();
				udpServerIsAlive = true;				
			}
			catch (Exception e)
			{
				MessageBox.Show("start_server() exception occurred : "+e.Message);
				
			}
		}
		
		/// <summary>
		///	This function will stop  and close the UDP server.  And
		/// it also aborts the running threads  
		/// </summary>
		public void stop_server()
		{
			try
			{	
				if (udpSocket != null)
				{
					udpSocket.Close();
				}
				udpListeningThread=null;//Free Mem
			}
			catch (Exception e)
			{
				MessageBox.Show("stop_server() exception occurred :"+e.Message);
			}
		}

		/// <summary>
		/// Receive UDP Packets
		/// </summary>
		public void receivePacket()
		{			
			IPHostEntry localHostEntry;
		
			try
			{
				udpSocket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
				localHostEntry = Dns.GetHostByName(Dns.GetHostName());
				IPEndPoint localIpEndPoint = new IPEndPoint(localHostEntry.AddressList[0], UDPCONSTANTS.UDPPORT);
				EndPoint localEndEntry = (localIpEndPoint);
				udpSocket.Bind(localEndEntry);
					
				while (udpServerIsAlive)
				{
					Monitor.Enter(this);
					Byte[] received = new Byte[UDPCONSTANTS.MESSAGESIZE];

					//This is just a dummy endpoint that will be replaced after the call to ReceiveFrom.
					//In the ReceiveFrom call later on, dummyIpEndPoint will be passed by ref.
					IPEndPoint dummyIpEndPoint = new IPEndPoint(localHostEntry.AddressList[0], UDPCONSTANTS.UDPPORT);

					EndPoint remoteNotifyEP = (dummyIpEndPoint);

					int udpBytesReceived = udpSocket.ReceiveFrom(received, ref remoteNotifyEP);
					//MessageBox.Show(udpBytesReceived.ToString());
					ReceivedDataLength=udpBytesReceived;								
					String dataReceivedUdp = System.Text.Encoding.ASCII.GetString(received,0,received.Length );
					ReceivedData=dataReceivedUdp.ToString();
					//HandleReceivedPacket();
					controlInvoker.Invoke(new MethodCallInvoker(HandleReceivedPacket));					
					Monitor.Exit(this);					
				}		
					
			}	
			catch (Exception e)
			{
				//MessageBox.Show("Exception occurred in receivedPacket() :"+e.Message);
			}
				
		}
		private void DisplayDebug(string p_data)
		{
			ChatBox.Text = ChatBox.Text+"\r\n"+"Debug: "+"\r\n"+p_data;
		}

		//UdpClient relate methods
		public bool sendUDP(string serverIPaddr, string message)
		{
			Byte[] receivedData = new Byte[UDPCONSTANTS.MESSAGESIZE];
			bool sentSuccess = false;
			try
			{
				IPAddress serverIP = IPAddress.Parse(serverIPaddr);
				udpClient = new UdpClient();
				udpClient.Connect(serverIP,UDPCONSTANTS.UDPPORT);
		
				Byte[] inputToBeSent = new Byte[UDPCONSTANTS.MESSAGESIZE];
		
				//Convert the "message" string to a byte array, and then send it to
				//the remote UDP server using udpClient.
				inputToBeSent = System.Text.Encoding.ASCII.GetBytes(message.ToCharArray());
				int nBytesSent = udpClient.Send(inputToBeSent, inputToBeSent.Length);
				sentSuccess = true;
				
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception is occurred in sendUDP() : "+e.Message);
			}
			udpClient.Close();
			return sentSuccess;
		}

		private Packet getPacketFromReceiptBuffer(string p_receiptID)
		{
			Packet packet = new Packet();
			IDictionaryEnumerator ide = ReceiptBuffer.GetEnumerator();
			while(ide.MoveNext())
			{
				if(ide.Key.ToString().Equals(p_receiptID))
				{
					packet = (Packet)ide.Value;
					break;
				}
			}
			return packet; 
		}
		/// <summary>
		/// Send a route error packet to the initiator.
		/// It does not have route reply packet back
		/// If source does not have any packet such as route reply packet 
		/// or route error packet, it will time out.
		/// </summary>
		public void InformRouteErrorToSource(Packet p_packet)
		{
			try
			{			
				if(p_packet.strType.Equals(Packet_CONSTANTS.DATA_PACKET) ||
					p_packet.strType.Equals(Packet_CONSTANTS.HANG_UP_PACKET) ||
					p_packet.strType.Equals(Packet_CONSTANTS.REPLY_HANG_UP_PACKET) ||
					p_packet.strType.Equals(Packet_CONSTANTS.ROUTE_REPLY_PACKET))
				{
					string p_receiptID = p_packet.strReceiptID;
					if(ReceiptBuffer.Contains(p_receiptID))
					{
						Packet packet = getPacketFromReceiptBuffer(p_receiptID);
						ReceiptBuffer.Remove(p_receiptID);
						SendRouteErrorPacket(packet);
					}
				}
			}//try
			catch(Exception e)
			{
				MessageBox.Show("Exception occurred in InformRouteErrorToSource() : "+e.Message);
			}

		}


		/// <summary>
		/// Send route error packet to the source.
		/// </summary>
		/// <param name="p_packet">packet which is not transmitted: got it from receipt buffer</param>
		private void SendRouteErrorPacket(Packet p_packet)
		{
			try
			{
				string originalType = p_packet.strType;
				string temp_RouteRecord = p_packet.strRouteRecord;
				p_packet.strType = Packet_CONSTANTS.ROUTE_ERROR_PACKET;                
				//string errorPacketID = p_packet.strRequestID;
				string source = "";

				if(originalType.Equals(Packet_CONSTANTS.DATA_PACKET) ||
					originalType.Equals(Packet_CONSTANTS.HANG_UP_PACKET))
				{
					source = p_packet.strInitiator;
				}
				else if(originalType.Equals(Packet_CONSTANTS.ROUTE_REPLY_PACKET) ||
					originalType.Equals(Packet_CONSTANTS.REPLY_HANG_UP_PACKET))//reply packets
				{
					source = p_packet.strTarget;
				}
				p_packet.strInitiator = strMyID;
				p_packet.strTarget = source;
				p_packet.strRouteRecord = strMyID;
				p_packet.strPayload = Packet_CONSTANTS.NOTAVAILABLE;
				p_packet.strCurrentID = strMyID;
				p_packet.strRouteCache = getReverseRoute(temp_RouteRecord);
				if(source.Equals(strMyID))//when route error ocurrs in source itself
				{
					string serverIP = GetIPAddressByIDInConfFile(source);					
					SaveInReceiptBuffer(p_packet);//for a receipt packet
					string messageToSend = CreateSendingMessage(p_packet);
					//showDebugSendingPackets(p_packet);
					sendUDP(serverIP, messageToSend);
				}
				else
				{
					SendPacket(ref p_packet);				
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SendRouteErrorPacket(): "+e.Message);
			}		
		}

		private string CreateSendingMessage(Packet p_packet)
		{
            string messageToSend = "NA";
			try
			{
				XmlDocument doc = new XmlDocument();
				doc.LoadXml("<Packet></Packet>");

				XmlNode root = doc.DocumentElement;

				//1. Type
				XmlElement elem = doc.CreateElement(Packet_CONSTANTS.TYPE);
				elem.InnerText = p_packet.strType;
				root.AppendChild(elem);


				//2.Initiator
				elem = doc.CreateElement(Packet_CONSTANTS.INITIATOR);
				elem.InnerText = p_packet.strInitiator;
				root.AppendChild(elem);


				//3.Current ID
				elem = doc.CreateElement(Packet_CONSTANTS.CURRENTID);
				elem.InnerText = p_packet.strCurrentID;
				root.AppendChild(elem);
				
				//4.Payload
				elem = doc.CreateElement(Packet_CONSTANTS.PAYLOAD);
				elem.InnerText = p_packet.strPayload;
				root.AppendChild(elem);

				//5. RequestID
				elem = doc.CreateElement(Packet_CONSTANTS.REQUESTID);
				elem.InnerText = p_packet.strRequestID;
				root.AppendChild(elem);

				//6.RouteCache
				elem = doc.CreateElement(Packet_CONSTANTS.ROUTECACHE);
				elem.InnerText = p_packet.strRouteCache;
				root.AppendChild(elem);

				//7.RouteRecord
				elem = doc.CreateElement(Packet_CONSTANTS.ROUTERECORD);
				elem.InnerText = p_packet.strRouteRecord;
				root.AppendChild(elem);

				//8.Target
				elem = doc.CreateElement(Packet_CONSTANTS.TARGET);
				elem.InnerText = p_packet.strTarget;
				root.AppendChild(elem);
				
				//9.ReceiptID
				elem = doc.CreateElement(Packet_CONSTANTS.RECEIPTID);
				elem.InnerText = p_packet.strReceiptID;
				root.AppendChild(elem);

				//10. Put TimeStamp into packet
				elem = doc.CreateElement(Packet_CONSTANTS.TIMESTAMP);
				//p_packet.strTimeStamp = NTTimer.getCount().ToString();
				p_packet.strTimeStamp = getNTTime().ToString();
				elem.InnerText = p_packet.strTimeStamp;
				root.AppendChild(elem);
				messageToSend = doc.InnerXml.ToString();
			}
			catch(Exception e)
			{
              MessageBox.Show("Exception in CreateSendingMessage(): "+e.Message);
			}
            return messageToSend;
		}

		private Packet getLostPacketFromSendBuffer(string p_requestID)
		{
			Packet packet = new Packet();
			try
			{
				IDictionaryEnumerator ide = send_buffer.GetEnumerator();
				while(ide.MoveNext())
				{
					if(p_requestID.Equals(ide.Key.ToString()))
					{
						packet = (Packet)ide.Value;
						break;
					}
				}
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception in getLostPacketFromSendBuffer():"+e.Message);
			}
			return packet;
		}

		//for debugging purpose
		private string insideSendBuffer()
		{
			string inside ="";
			try
			{
				IDictionaryEnumerator ide = send_buffer.GetEnumerator();
				while(ide.MoveNext())
				{
					inside = inside + ide.Key.ToString()+": ";
					Packet temp_packet = (Packet)ide.Value;
					inside = inside + temp_packet.strType+"\r\n";
				}
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception in insideSendBuffer(): "+e.Message); 
			}
			return inside;
		}

		/// <summary>
		/// this method is performed in the source node, which has received route error packet.
		/// </summary>
		/// <param name="p_packet"></param>
		public void PerformRouteDiscovery(Packet p_packet)
		{
			try
			{
				if(send_buffer.Contains(p_packet.strRequestID) &&
					!processingRDBuffer.Contains(p_packet.strRequestID))
				{					
					processingRDBuffer.Add(p_packet.strRequestID);
					Packet packet = new Packet();
					packet.strType = Packet_CONSTANTS.ROUTE_DISCOVERY_PACKET;
					packet.strInitiator = strMyID; //which is source itself
					packet.strTarget = p_packet.strTarget;					
					packet.strRequestID = getRequestID(Packet_CONSTANTS.ROUTE_DISCOVERY_PACKET);
					packet.strCurrentID = strMyID;
					//by using payload slot, put the lost packet's request id.
					packet.strPayload = p_packet.strRequestID;
					packet.strRouteCache = Packet_CONSTANTS.NOTAVAILABLE;
					packet.strRouteRecord = strMyID;				
					SaveInSendBuffer(packet);
					string timerID = packet.strRequestID;
					if(!RRTimer2Buffer.Contains(timerID))
					{
						SetRouteReplyTimer(timerID, packet);
						SendPacket(ref packet);				
					}					
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception occurred in PerformRouteDiscovery():"+e.Message);
			}
		}

		/// <summary>
		/// This repairing process should be done in source node.
		/// If route are available in source's cache, source uses it. 
		/// Otherwise, perform route discovery.
		/// </summary>
		/// <param name="packet"></param>
		public void RepairRouteError(ref Packet packet)
		{
			try
			{
				if(send_buffer.Contains(packet.strRequestID))
				{
					string requestID = packet.strRequestID;
					Packet sb_packet = getLostPacketFromSendBuffer(requestID);
					if(sb_packet.strInitiator.Equals(strMyID))
					{						
						//ChatBox.Text = ChatBox.Text+"\r\n"+"Route Error Repair in : "+strMyID;
						if(!routeCache.isEmpty())
						{
							//ChatBox.Text += "\r\n"+"Route Error Repair Using another Route in Cache";
							sb_packet.strRouteCache = routeCache.getPath();
							string requestIDSB = sb_packet.strRequestID;
							send_buffer.Remove(requestIDSB);
							send_buffer.Add(requestIDSB, sb_packet);
							if(!RRTimer2Buffer.Contains(sb_packet.strRequestID))
							{
								SetRouteReplyTimer(sb_packet.strRequestID, sb_packet);
								SendPacket(ref sb_packet);		
							}
						}
						else
						{
							//ChatBox.Text = ChatBox.Text+"\r\n"+"Repair Route Error will perform route discovery";
							PerformRouteDiscovery(sb_packet);
						}                 
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception ocurred in RepairRouteError():"+e.Message);
			}
		}

		private void showNeighborsInside()
		{
			try
			{
				IDictionaryEnumerator ide = source_neighbors.GetEnumerator();
				string inside = "";
				while(ide.MoveNext())
				{
					inside = inside + ide.Key.ToString()+" : ";
					ArrayList al = (ArrayList)ide.Value;
					IEnumerator ie = al.GetEnumerator();
					while(ie.MoveNext())
					{
						inside = inside + ie.Current.ToString()+" ";    
					}
					inside = inside+"\r\n";
				}
				MessageBox.Show("NeighborInside"+"\r\n"+inside);			   
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in showNeighborsInside() : "+e.Message);
			}
	  }

		private void BtnNT_Click(object sender, System.EventArgs e)
		{
			try
			{
				MessageBox.Show("NTTime: "+getNTTime().ToString());
			}
			catch(Exception ex)
			{
				MessageBox.Show("Exception in BtnNT_Click() : "+ex.Message);
			}
		}


		/// <summary>
		/// Set ACM on or off
		/// </summary>
		/*private void SetACM()
		{
			try
			{
				if(ACM_CB.Checked)
				{
					OnACM = true;
				}
				else
				{
					OnACM = false;
				}
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in SetACM(): "+e.Message);
			}
		}*/

		/*private void SetCache()
		{
			try
			{
				if(Cache_CB.Checked)
				{
					OnCache = true;
				}
				else
				{
                    OnCache = false;
				}
			}
			catch(Exception e)
			{
               MessageBox.Show("Exception in SetCache(): "+e.Message);
			}
		}*/

		private void Start_Btn_Click(object sender, System.EventArgs e)
		{
			showBuddyScreen();
			resetAll();									
		}

		private void Conf_CB_CheckStateChanged(object sender, System.EventArgs e)
		{
		    UseOfConf = Conf_CB.Checked;
		}




	}//main class

	public struct ElementsForReceiverWindow
	{
		public string SN;
		public string Payload;
	}

	public struct BroadcastPacket
	{
		public string strType;		
		//public string strOnCache;
		//public string strOnACM;
		//public string strSWS;
		//public string strOnRP;
		public string strOnConf;
	}

	/// <summary>
	/// This is a struct for Packet
	/// </summary>
	public struct Packet
	{
		public string strType;
		public string strInitiator;
		public string strTarget;
		public string strRequestID;
		public string strRouteRecord;
		public string strCurrentID;
		public string strPayload;
		public string strRouteCache;
		public string strReceiptID;
		public string strTimeStamp;

		public bool isDataPacket()
		{
			if(strType.Equals(Packet_CONSTANTS.DATA_PACKET))
			{				
				return true;
			}
			else
			{
                 return false;
			}
		}
	}

	public struct ReceiptPacket
	{
          public string strType;
		  public string strReceiptID;
	}

	/// <summary>
	/// one path in one sequence consists of three: destination, speed, and pause
	/// </summary>
	public struct Sequence
	{
		public Point destination;
		public int speed;
		public int pause;

	}

	public struct Point
	{
		public double row;
		public double col;
	
		public Point(double p_row, double p_col)
		{
			row = Convert.ToDouble(Math.Round(Convert.ToDecimal(p_row),2));
			col = Convert.ToDouble(Math.Round(Convert.ToDecimal(p_col),2));
		}

	}//end of struct
	
	public class MyTimer2 
	{
		//properties	
		private Thread thread;
		private int _interval;
		private bool _enabled;
		private int _count;
		private int _condition;
		private PocketPCChat _pc;
		private string _ID;
		private Packet _packet;
		private int _numOfCall;
		public string NTCurrentThread;

		public MyTimer2()
		{
			_count = 0;
			_numOfCall = 0;
			_interval = 100;
			_enabled = false;	
			_condition = -1;				
		}

		public MyTimer2(int p_interval, bool p_enabled)
		{
             _count = 0;
			 _numOfCall = 0;
			 _interval = p_interval;
			 _enabled = p_enabled;
			 _condition = -1;
		}

		public MyTimer2(int p_interval, bool p_enabled, PocketPCChat p_pc)
		{
            _interval = p_interval;
			_enabled = p_enabled;
			_pc = p_pc;
		}
		public MyTimer2(int p_interval, bool p_enabled, int p_condition, 
			PocketPCChat p_pc, Packet p_packet, string p_ID)
		{
			_count = 0;
			_numOfCall = 0;
			_interval = p_interval;
			_enabled = p_enabled;
            _condition = p_condition;
			_pc = p_pc;
			_packet = p_packet;
			_ID = p_ID;
		}

		public void setEnabled(bool p_enabled)
		{
			_enabled = p_enabled;				 
		}

		public void setInterval(int p_interval)
		{
			_interval = p_interval;
		}

		public int getInterval()
		{
			return _interval;
		}

		public bool getEnabled()
		{
			return _enabled;
		}

		private void NTTimer_Click()
		{
			try
			{
				while(_enabled)
				{
					_count++;
					NTCurrentThread = Thread.CurrentThread.GetHashCode().ToString();
					
					/*if(_count==1 || _count==2)
					{
						MessageBox.Show("NTTimer: "+NTCurrentThread+" "+_count);
					}*/
					Thread.Sleep(_interval);
				}//while
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception in NTTimer_Click(): "+e.Message);
			}
		}

		public int getCount()
		{
			return _count;
		}
		public void setCondition(int p_condition)
		{
			_condition = p_condition;
		}

		public void resetCount()
		{
			_count = 0;
		}

		public void NTTimer_Start()
		{
			thread = new Thread(new ThreadStart(this.NTTimer_Click));
			thread.Start();			
		}

		private void RRTimer2_Click()
		{
			try
			{				
				while(_enabled || _pc.send_buffer.Contains(_packet.strRequestID))
				{
					Monitor.Enter(this);
					Thread.Sleep(_interval);
    				_count++;
						if(_enabled)
						{
							if(_count<_condition)
							{
								if(_packet.strType.Equals(Packet_CONSTANTS.DATA_PACKET) ||
									_packet.strType.Equals(Packet_CONSTANTS.HANG_UP_PACKET) ||
									_packet.strType.Equals(Packet_CONSTANTS.START_CHAT_PACKET))
								{
									if(_pc.send_buffer.Contains(_packet.strRequestID))
									{
										_pc.SendPacket(ref _packet);
									}
								}
								else //route discovery packet
								{
									string lostPacketRequestID = _packet.strPayload;
									if(_pc.send_buffer.Contains(lostPacketRequestID) && 
										_pc.send_buffer.Contains(_packet.strRequestID))
									{
										_pc.SendPacket(ref _packet);
									}
								}
							}
							else 
							{
								_enabled = false;
								resetCount();
								_pc.ReleaseRRTimer(_packet.strRequestID);
								if(_packet.strType.Equals(Packet_CONSTANTS.DATA_PACKET) ||
									_packet.strType.Equals(Packet_CONSTANTS.HANG_UP_PACKET))
								{
									if(_pc.send_buffer.Contains(_packet.strRequestID))
									{
										//MessageBox.Show("After three trials, repair old route: "+_packet.strRouteCache+" with type of "+
										//	_packet.strType+"by using RepairRouteError()");
										_pc.routeCache.RemovePath(_packet.strRouteCache);
										_pc.RepairRouteError(ref _packet);
									}
								}
								else //Start Chat Packet, Route Discovery
								{
									if(_packet.strType.Equals(Packet_CONSTANTS.START_CHAT_PACKET))
									{
										if(_pc.send_buffer.Contains(_packet.strRequestID))
										{
											_pc.routeCache.RemovePath(_packet.strRouteCache);
											_pc.send_buffer.Remove(_packet.strRequestID);//RD packet
											MessageBox.Show("Can't Find Any Route to "+_packet.strTarget, "No Route Found",
												MessageBoxButtons.OK, MessageBoxIcon.Exclamation, MessageBoxDefaultButton.Button1);
										}
									}
									else//Route Discovery Packet
									{
										if(_pc.send_buffer.Contains(_packet.strRequestID))
										{
											_pc.routeCache.RemovePath(_packet.strRouteCache);
											_pc.send_buffer.Remove(_packet.strRequestID);//RD packet
											_pc.send_buffer.Remove(_packet.strPayload);//data packet
											_pc.SenderWindow.Remove(_packet.strPayload);//remove data packet in SenderWindow
											MessageBox.Show("Can't Find Any Route to "+_packet.strTarget, "No Route Found",
												MessageBoxButtons.OK, MessageBoxIcon.Exclamation, MessageBoxDefaultButton.Button1);
										}
									}
								}
								break;
							}
						}//check again
					//}//when appropriate time is elapsed
                    Monitor.Exit(this);
				}//while
				
			}
			catch(Exception e)
			{
                MessageBox.Show("RRTimer2_Click(): "+e.Message);
			}
		}

		public void SetRRTimer2()
		{
			try
			{
				thread = new Thread(new ThreadStart(this.RRTimer2_Click));
				thread.Start();
			}
			catch(Exception e)
			{
                MessageBox.Show("Exception in SetRRTimer2(): "+e.Message);
			}
		}

		public void SetReceiptTimer2()
		{			
            thread = new Thread(new ThreadStart(this.ReceiptTimer2_Click));
			thread.Start();
		}

		private void MakeCurrentThreadSleep()
		{
            int count = 0;
			while(500*count <= _interval)
			{
			   count++;
               Thread.Sleep(500); 
			}
   		}
		private void ReceiptTimer2_Click()
		{
			try
			{
				_numOfCall++;
				while(_enabled)
				{				
                    Monitor.Enter(this);
		//			if(isElapsedTimeEqualToInterval())
		//			{
					Thread.Sleep(_interval);
					//MakeCurrentThreadSleep();
					    //Thread.Sleep(_interval);
						if((_numOfCall == 1) && (_enabled))
						{
							_pc.ReleaseReceiptTimer(_ID);
							if(_pc.ReceiptBuffer.Contains(_ID))
							{
								_pc.InformRouteErrorToSource(_packet);
							}
							break;
						}												
//					}
					Monitor.Exit(this);
 				}
                _numOfCall = 0;
			}
			catch(Exception e)
			{
                MessageBox.Show("ReceiptTimer2_Click(): "+e.Message);
			}
		}

	}//inner class

	

}
